# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Data


class Labels(Data):
    """
    Tokenizes text-classification datasets as input for training text-classification models.
    """

    def __init__(self, tokenizer, columns, maxlength):
        """
        Creates a new instance for tokenizing Labels training data.

        Args:
            tokenizer: model tokenizer
            columns: tuple of columns to use for text/label
            maxlength: maximum sequence length
        """

        super().__init__(tokenizer, columns, maxlength)

        if not self.columns:
            self.columns = ("text", None, "label")
        elif len(columns) < 3:
            self.columns = (self.columns[0], None, self.columns[-1])

    def process(self, data):
        text1, text2, label = self.columns

        text = (data[text1], data[text2]) if text2 else (data[text1],)

        inputs = self.tokenizer(*text, max_length=self.maxlength, padding=True, truncation=True)
        inputs[label] = data[label]

        return inputs



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tarfile

from .compress import Compress


class Tar(Compress):
    """
    Tar compression
    """

    def pack(self, path, output):
        compression = self.compression(output)

        with tarfile.open(output, f"w:{compression}" if compression else "w") as tar:
            tar.add(path, arcname=".")

    def unpack(self, path, output):
        compression = self.compression(path)

        with tarfile.open(path, f"r:{compression}" if compression else "r") as tar:
            for member in tar.getmembers():
                fullpath = os.path.join(path, member.name)
                if not self.validate(path, fullpath):
                    raise OSError(f"Invalid tar entry: {member.name}")

            tar.extractall(output)

    def compression(self, path):
        """
        Gets compression type for path.

        Args:
            path: path to file

        Returns:
            compression type
        """

        compression = path.lower().split(".")[-1]
        return compression if compression in ("bz2", "gz", "xz") else None



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .audio import AudioMixer, AudioStream, Microphone, Signal, TextToAudio, TextToSpeech, Transcription
from .base import Pipeline
from .data import FileToHTML, HTMLToMarkdown, Segmentation, Tabular, Textractor, Tokenizer
from .factory import PipelineFactory
from .hfmodel import HFModel
from .hfpipeline import HFPipeline
from .image import ImageHash, Objects
from .nop import Nop
from .tensors import Tensors
from .text import CrossEncoder, Entity, Labels, Questions, Similarity, Summary, Translation

__all__ = (
    "AudioMixer",
    "AudioStream",
    "CrossEncoder",
    "Entity",
    "FileToHTML",
    "HFModel",
    "HFPipeline",
    "HTMLToMarkdown",
    "ImageHash",
    "Labels",
    "Microphone",
    "Nop",
    "Objects",
    "Pipeline",
    "PipelineFactory",
    "Questions",
    "Segmentation",
    "Signal",
    "Similarity",
    "Summary",
    "Tabular",
    "Tensors",
    "TextToAudio",
    "TextToSpeech",
    "Textractor",
    "Tokenizer",
    "Transcription",
    "Translation",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import os
import tempfile
from urllib.parse import urlparse
from urllib.request import Request, urlopen

from .filetohtml import FileToHTML
from .htmltomd import HTMLToMarkdown
from .segmentation import Segmentation


class Textractor(Segmentation):
    """
    Extracts text from files.
    """

    # pylint: disable=R0913
    def __init__(
        self,
        sentences=False,
        lines=False,
        paragraphs=False,
        minlength=None,
        join=False,
        sections=False,
        cleantext=True,
        chunker=None,
        headers=None,
        backend="available",
        **kwargs,
    ):
        super().__init__(sentences, lines, paragraphs, minlength, join, sections, cleantext, chunker, **kwargs)

        backend = "tika" if kwargs.get("tika") else None if "tika" in kwargs else backend

        self.html = FileToHTML(backend) if backend else None

        self.markdown = HTMLToMarkdown(self.paragraphs, self.sections)

        self.headers = headers if headers else {}

    def text(self, text):
        path, exists = self.valid(text)

        if not path:
            html = text

        elif self.html:
            path = path if exists else self.download(path)

            html = self.html(path)

            html = html if html else self.retrieve(path)

            if not exists:
                os.remove(path)

        else:
            html = self.retrieve(path)

        return self.markdown(html)

    def valid(self, path):
        """
        Checks if path is a valid local file or web url. Returns path if valid along with a flag
        denoting if the path exists locally.

        Args:
            path: path to check

        Returns:
            (path, exists)
        """

        path = path.replace("file://", "")

        exists = os.path.exists(path)

        return (path if exists or urlparse(path).scheme in ("http", "https") else None, exists)

    def download(self, url):
        """
        Downloads content of url to a temporary file.

        Args:
            url: input url

        Returns:
            temporary file path
        """

        with tempfile.NamedTemporaryFile(mode="wb", delete=False) as output:
            path = output.name

            output.write(self.retrieve(url))

        return path

    def retrieve(self, url):
        """
        Retrieves content from url.

        Args:
            url: input url

        Returns:
            data
        """

        if os.path.exists(url):
            with open(url, "rb") as f:
                return f.read()

        with contextlib.closing(urlopen(Request(url, headers=self.headers))) as connection:
            return connection.read()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import uuid


class AutoId:
    """
    Generates unique ids.
    """

    def __init__(self, method=None):
        """
        Creates a unique id generator.

        Args:
            method: generation method - supports int sequence (default) or UUID function
        """

        self.method, self.function, self.value = None, None, None

        if not method or isinstance(method, int):
            self.method = self.sequence
            self.value = method if method else 0
        else:
            self.method = self.uuid
            self.function = getattr(uuid, method)

        args = inspect.getfullargspec(self.function).args if self.function else []
        self.deterministic = "namespace" in args

    def __call__(self, data=None):
        """
        Generates a unique id.

        Args:
            data: optional data to use for deterministic algorithms (i.e. uuid3, uuid5)

        Returns:
            unique id
        """

        return self.method(data)

    # pylint: disable=W0613
    def sequence(self, data):
        """
        Gets and increments sequence.

        Args:
            data: not used

        Returns:
            current sequence value
        """

        value = self.value
        self.value += 1

        return value

    def uuid(self, data):
        """
        Generates a UUID and return as a string.

        Args:
            data: used with determistic algorithms (uuid3, uuid5)

        Returns:
            UUID string
        """

        uid = self.function(uuid.NAMESPACE_DNS, str(data)) if self.deterministic else self.function()
        return str(uid)

    def current(self):
        """
        Get the current sequence value. Only applicable for sequence ids, will be None for UUID methods.

        Returns:
            current sequence value
        """

        return self.value



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Pipeline:
    """
    Base class for all Pipelines. The only interface requirement is to define a __call___ method.
    """

    def batch(self, data, size):
        """
        Splits data into separate batch sizes specified by size.

        Args:
            data: data elements
            size: batch size

        Returns:
            list of evenly sized batches with the last batch having the remaining elements
        """

        return [data[x : x + size] for x in range(0, len(data), size)]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utilities import Resolver
from .external import External
from .huggingface import HFVectors
from .litellm import LiteLLM
from .sbert import STVectors
from .words import WordVectors


class VectorsFactory:
    """
    Methods to create dense vector models.
    """

    @staticmethod
    def create(config, scoring=None, models=None):
        """
        Create a Vectors model instance.

        Args:
            config: vector configuration
            scoring: scoring instance
            models: models cache

        Returns:
            Vectors
        """

        method = VectorsFactory.method(config)

        if method == "external":
            return External(config, scoring, models)

        if method == "litellm":
            return LiteLLM(config, scoring, models)

        if method == "sentence-transformers":
            return STVectors(config, scoring, models) if config and config.get("path") else None

        if method == "words":
            return WordVectors(config, scoring, models)

        if HFVectors.ismethod(method):
            return HFVectors(config, scoring, models) if config and config.get("path") else None

        return VectorsFactory.resolve(method, config, scoring, models) if method else None

    @staticmethod
    def method(config):
        """
        Get or derive the vector method.

        Args:
            config: vector configuration

        Returns:
            vector method
        """

        method = config.get("method")
        path = config.get("path")

        if not method:
            if path:
                if LiteLLM.ismodel(path):
                    method = "litellm"
                elif WordVectors.ismodel(path):
                    method = "words"
                else:
                    method = "transformers"
            elif config.get("transform"):
                method = "external"

        return method

    @staticmethod
    def resolve(backend, config, scoring, models):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: vector configuration
            scoring: scoring instance
            models: models cache

        Returns:
            Vectors
        """

        try:
            return Resolver()(backend)(config, scoring, models)
        except Exception as e:
            raise ImportError(f"Unable to resolve vectors backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch


class Tokens(torch.utils.data.Dataset):
    """
    Default dataset used to hold tokenized data.
    """

    def __init__(self, columns):
        self.data = []

        for column in columns:
            for x, value in enumerate(columns[column]):
                if len(self.data) <= x:
                    self.data.append({})

                self.data[x][column] = value

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import FunctionType, MethodType


class Functions:
    """
    Resolves function configuration to function references.
    """

    def __init__(self, embeddings):
        """
        Creates a new function resolver.

        Args:
            embeddings: embeddings instance
        """

        self.embeddings = embeddings

        self.references = None

    def __call__(self, config):
        """
        Resolves a list of functions to function references.

        Args:
            config: configuration

        Returns:
            list of function references
        """

        self.references = []

        functions = []
        for fn in config["functions"]:
            if isinstance(fn, dict):
                fn = fn.copy()
                fn["function"] = self.function(fn["function"])
            else:
                fn = self.function(fn)
            functions.append(fn)

        return functions

    def reset(self):
        """
        Clears all resolved references.
        """

        if self.references:
            for reference in self.references:
                reference.reset()

    def function(self, function):
        """
        Resolves function configuration. If function is a string, it's split on '.' and each part
        is separately resolved to an object, attribute or function. Each part is resolved upon the
        first invocation of the function. Otherwise, the input is returned.

        Args:
            function: function configuration

        Returns:
            function reference
        """

        if isinstance(function, str):
            parts = function.split(".")

            if hasattr(self.embeddings, parts[0]):
                m = Reference(self.embeddings, parts[0])
                self.references.append(m)
            else:
                module = ".".join(parts[:-1])
                m = __import__(module)

            for comp in parts[1:]:
                m = Reference(m, comp)
                self.references.append(m)

            return m

        return function


class Reference:
    """
    Stores a reference to an object attribute. This attribute is resolved by invoking the __call__ method.
    This allows for functions to be independent of the initialization order of an embeddings instance.
    """

    def __init__(self, obj, attribute):
        """
        Create a new reference.

        Args:
            obj: object handle
            attribute: attribute name
        """

        self.obj = obj
        self.attribute = attribute

        self.inputs = (obj, attribute)

        self.resolved = False

        self.function = None

    def __call__(self, *args):
        """
        Resolves an object attribute reference. If the attribute is a function, the function is executed.
        Otherwise, the object attribute value is returned.

        Args:
            args: list of function arguments to the object attribute, when attribute is a function

        Returns:
            object attribute function result or object attribute value
        """

        if not self.resolved:
            self.obj = self.obj() if isinstance(self.obj, Reference) else self.obj
            self.attribute = self.attribute() if isinstance(self.attribute, Reference) else self.attribute
            self.resolved = True

        attribute = getattr(self.obj, self.attribute)

        if self.function is None:
            self.function = isinstance(attribute, FunctionType | MethodType) or (callable(attribute) and args)

        return attribute(*args) if self.function else attribute

    def reset(self):
        """
        Clears resolved references.
        """

        self.obj, self.attribute = self.inputs
        self.resolved = False



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os

try:
    import pandas as pd

    PANDAS = True
except ImportError:
    PANDAS = False

from .base import Task


class ExportTask(Task):
    """
    Task that exports task elements using Pandas.
    """

    def register(self, output=None, timestamp=None):
        """
        Add export parameters to task. Checks if required dependencies are installed.

        Args:
            output: output file path
            timestamp: true if output file should be timestamped
        """

        if not PANDAS:
            raise ImportError('ExportTask is not available - install "workflow" extra to enable')

        # pylint: disable=W0201
        self.output = output
        self.timestamp = timestamp

    def __call__(self, elements, executor=None):
        outputs = super().__call__(elements, executor)

        output = self.output
        parts = list(os.path.splitext(output))
        extension = parts[-1].lower()

        if self.timestamp:
            timestamp = datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%dT%H%M%SZ")
            parts[-1] = timestamp + parts[-1]

            output = ".".join(parts)

        if extension == ".xlsx":
            pd.DataFrame(outputs).to_excel(output, index=False)
        else:
            pd.DataFrame(outputs).to_csv(output, index=False)

        return outputs



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

try:
    from nltk import sent_tokenize

    NLTK = True
except ImportError:
    NLTK = False


try:
    import chonkie  # type:ignore

    CHONKIE = True
except ImportError:
    CHONKIE = False

from ..base import Pipeline


class Segmentation(Pipeline):
    """
    Segments text into logical units.
    """

    def __init__(
        self,
        sentences=False,
        lines=False,
        paragraphs=False,
        minlength=None,
        join=False,
        sections=False,
        cleantext=True,
        chunker=None,
        **kwargs,
    ):
        """
        Creates a new Segmentation pipeline.

        Args:
            sentences: tokenize text into sentences if True, defaults to False
            lines: tokenizes text into lines if True, defaults to False
            paragraphs: tokenizes text into paragraphs if True, defaults to False
            minlength: require at least minlength characters per text element, defaults to None
            join: joins tokenized sections back together if True, defaults to False
            sections: tokenizes text into sections if True, defaults to False. Splits using section or page breaks,
                        depending on what's available
            cleantext: apply text cleaning rules, defaults to True
            chunker: creates a third-party chunker to tokenize text if set, defaults to None
            kwargs: additional keyword arguments
        """

        if not NLTK and sentences:
            raise ImportError('NLTK is not available - install "pipeline" extra to enable')

        if not CHONKIE and chunker:
            raise ImportError('Chonkie is not available - install "pipeline" extra to enable')

        self.sentences = sentences
        self.lines = lines
        self.paragraphs = paragraphs
        self.sections = sections
        self.minlength = minlength
        self.join = join
        self.cleantext = cleantext

        self.chunker = self.createchunker(chunker, **kwargs) if chunker else None

    def __call__(self, text):
        """
        Segments text into semantic units.

        This method supports text as a string or a list. If the input is a string, the return
        type is text|list. If text is a list, a list of returned, this could be a
        list of text or a list of lists depending on the tokenization strategy.

        Args:
            text: text|list

        Returns:
            segmented text
        """

        texts = [text] if not isinstance(text, list) else text

        results = []
        for value in texts:
            value = self.text(value)

            results.append(self.parse(value))

        return results[0] if isinstance(text, str) else results

    def text(self, text):
        """
        Hook to allow extracting text out of input text object.

        Args:
            text: object to extract text from
        """

        return text

    def parse(self, text):
        """
        Splits and cleans text based on the current parameters.

        Args:
            text: input text

        Returns:
            parsed and clean content
        """

        content = None

        if self.chunker:
            # pylint: disable=E1102
            content = [self.clean(x.text) for x in self.chunker(text)]
        elif self.sentences:
            content = [self.clean(x) for x in sent_tokenize(text)]
        elif self.lines:
            content = [self.clean(x) for x in re.split(r"\n{1,}", text)]
        elif self.paragraphs:
            content = [self.clean(x) for x in re.split(r"\n{2,}", text)]
        elif self.sections:
            split = r"\f" if "\f" in text else r"\n{3,}"
            content = [self.clean(x) for x in re.split(split, text)]
        else:
            content = self.clean(text)

        if isinstance(content, list):
            content = [x for x in content if x]
            return " ".join(content) if self.join else content

        return content

    def clean(self, text):
        """
        Applies a series of rules to clean text.

        Args:
            text: input text

        Returns:
            clean text
        """

        if not self.cleantext:
            return text

        text = re.sub(r" +", " ", text)
        text = text.strip()

        return text if not self.minlength or len(text) >= self.minlength else None

    def createchunker(self, chunker, **kwargs):
        """
        Creates a new third-party chunker

        Args:
            chunker: name of chunker to create
            kwargs: additional keyword arguments

        Returns:
            new chunker
        """

        chunker = f"{chunker[0].upper() + chunker[1:]}Chunker"
        return getattr(chonkie, chunker)(**kwargs)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import numpy as np

try:
    from pgvector.sqlalchemy import BIT, HALFVEC, VECTOR
    from sqlalchemy import Column, Index, Integer, MetaData, StaticPool, Table, create_engine, delete, func, text
    from sqlalchemy.orm import Session
    from sqlalchemy.schema import CreateSchema

    PGVECTOR = True
except ImportError:
    PGVECTOR = False

from ..base import VectoreIndex


class PGVector(VectoreIndex):
    """
    Builds an VectoreIndex index backed by a Postgres database.
    """

    def __init__(self, config):
        super().__init__(config)

        if not PGVECTOR:
            raise ImportError('PGVector is not available - install "ann" extra to enable')

        self.engine, self.database, self.connection, self.table = None, None, None, None

        quantize = self.config.get("quantize")
        self.qbits = quantize if quantize and isinstance(quantize, int) and not isinstance(quantize, bool) else None

    def load(self, path):
        self.initialize()

    def index(self, embeddings):
        self.initialize(recreate=True)

        self.database.execute(
            self.table.insert(), [{"indexid": x, "embedding": self.prepare(row)} for x, row in enumerate(embeddings)]
        )

        self.config["offset"] = embeddings.shape[0]
        self.metadata(self.settings())

    def append(self, embeddings):
        self.database.execute(
            self.table.insert(),
            [{"indexid": x + self.config["offset"], "embedding": self.prepare(row)} for x, row in enumerate(embeddings)],
        )

        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        self.database.execute(delete(self.table).where(self.table.c["indexid"].in_(ids)))

    def search(self, queries, limit):
        results = []
        for query in queries:
            query = self.database.query(self.table.c["indexid"], self.query(query)).order_by("score").limit(limit)

            results.append([(indexid, self.score(score)) for indexid, score in query])

        return results

    def count(self):
        # pylint: disable=E1102
        return self.database.query(func.count(self.table.c["indexid"])).scalar()

    def save(self, path):
        self.database.commit()
        self.connection.commit()

    def close(self):
        super().close()

        if self.database:
            self.database.close()
            self.engine.dispose()

    def initialize(self, recreate=False):
        """
        Initializes a new database session.

        Args:
            recreate: Recreates the database tables if True
        """

        self.connect()

        self.schema()

        table = self.setting("table", self.defaulttable())

        column, index = self.column()

        self.table = Table(
            table,
            MetaData(),
            Column("indexid", Integer, primary_key=True, autoincrement=False),
            Column("embedding", column),
        )

        index = Index(
            f"{table}-index",
            self.table.c["embedding"],
            postgresql_using="hnsw",
            postgresql_with=self.settings(),
            postgresql_ops={"embedding": index},
        )

        if recreate:
            self.table.drop(self.connection, checkfirst=True)
            index.drop(self.connection, checkfirst=True)

        self.table.create(self.connection, checkfirst=True)
        index.create(self.connection, checkfirst=True)

    def connect(self):
        """
        Establishes a database connection. Cleans up any existing database connection first.
        """

        if self.database:
            self.close()

        self.engine = create_engine(self.url(), poolclass=StaticPool, echo=False)
        self.connection = self.engine.connect()

        self.database = Session(self.connection)

        self.sqldialect(text("CREATE EXTENSION IF NOT EXISTS vector"))

    def schema(self):
        """
        Sets the database schema, if available.
        """

        schema = self.setting("schema")
        if schema:
            with self.engine.begin():
                self.sqldialect(CreateSchema(schema, if_not_exists=True))

            self.sqldialect(text("SET search_path TO :schema,public"), {"schema": schema})

    def settings(self):
        """
        Returns settings for this index.

        Returns:
            dict
        """

        return {"m": self.setting("m", 16), "ef_construction": self.setting("efconstruction", 200)}

    def sqldialect(self, sql, parameters=None):
        """
        Executes a SQL statement based on the current SQL dialect.

        Args:
            sql: SQL to execute
            parameters: optional bind parameters
        """

        args = (sql, parameters) if self.engine.dialect.name == "postgresql" else (text("SELECT 1"),)
        self.database.execute(*args)

    def defaulttable(self):
        """
        Returns the default table name.

        Returns:
            default table name
        """

        return "vectors"

    def url(self):
        """
        Reads the database url parameter.

        Returns:
            database url
        """

        return self.setting("url", os.environ.get("ANN_URL"))

    def column(self):
        """
        Gets embedding column and index definitions for the current settings.

        Returns:
            embedding column definition, index definition
        """

        if self.qbits:
            return BIT(self.config["dimensions"] * 8), "bit_hamming_ops"

        if self.setting("precision") == "half":
            return HALFVEC(self.config["dimensions"]), "halfvec_ip_ops"

        return VECTOR(self.config["dimensions"]), "vector_ip_ops"

    def prepare(self, data):
        """
        Prepares data for the embeddings column. This method returns a bit string for bit vectors and
        the input data unmodified for float vectors.

        Args:
            data: input data

        Returns:
            data ready for the embeddings column
        """

        if self.qbits:
            return "".join(np.where(np.unpackbits(data), "1", "0"))

        return data

    def query(self, query):
        """
        Creates a query statement from an input query. This method uses hamming distance for bit vectors and
        the max_inner_product for float vectors.

        Args:
            query: input query

        Returns:
            query statement
        """

        query = self.prepare(query)

        if self.qbits:
            return self.table.c["embedding"].hamming_distance(query).label("score")

        return self.table.c["embedding"].max_inner_product(query).label("score")

    def score(self, score):
        """
        Calculates the index score from the input score. This method returns the hamming score
        (1.0 - (hamming distance / total number of bits)) for bit vectors and the -score for
        float vectors.

        Args:
            score: input score

        Returns:
            index score
        """

        if self.qbits:
            return min(max(0.0, 1.0 - (score / (self.config["dimensions"] * 8))), 1.0)

        return -score



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Scoring:
    """
    Base scoring.
    """

    def __init__(self, config=None):
        """
        Creates a new Scoring instance.

        Args:
            config: input configuration
        """

        self.config = config if config is not None else {}

        columns = self.config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        self.model = None

    def insert(self, documents, index=None, checkpoint=None):
        """
        Inserts documents into the scoring index.

        Args:
            documents: list of (id, dict|text|tokens, tags)
            index: indexid offset
            checkpoint: optional checkpoint directory, enables indexing restart
        """

        raise NotImplementedError

    def delete(self, ids):
        """
        Deletes documents from scoring index.

        Args:
            ids: list of ids to delete
        """

        raise NotImplementedError

    def index(self, documents=None):
        """
        Indexes a collection of documents using a scoring method.

        Args:
            documents: list of (id, dict|text|tokens, tags)
        """

        if documents:
            self.insert(documents)

    def upsert(self, documents=None):
        """
        Convience method for API clarity. Calls index method.

        Args:
            documents: list of (id, dict|text|tokens, tags)
        """

        self.index(documents)

    def weights(self, tokens):
        """
        Builds a weights vector for each token in input tokens.

        Args:
            tokens: input tokens

        Returns:
            list of weights for each token
        """

        raise NotImplementedError

    def search(self, query, limit=3):
        """
        Search index for documents matching query.

        Args:
            query: input query
            limit: maximum results

        Returns:
            list of (id, score) or (data, score) if content is enabled
        """

        raise NotImplementedError

    def batchsearch(self, queries, limit=3, threads=True):
        """
        Search index for documents matching queries.

        Args:
            queries: queries to run
            limit: maximum results
            threads: run as threaded search if True and supported
        """

        raise NotImplementedError

    def count(self):
        """
        Returns the total number of documents indexed.

        Returns:
            total number of documents indexed
        """

        raise NotImplementedError

    def load(self, path):
        """
        Loads a saved Scoring object from path.

        Args:
            path: directory path to load scoring index
        """

        raise NotImplementedError

    def save(self, path):
        """
        Saves a Scoring object to path.

        Args:
            path: directory path to save scoring index
        """

        raise NotImplementedError

    def close(self):
        """
        Closes this Scoring object.
        """

        raise NotImplementedError

    def findmodel(self):
        """
        Returns the associated vector model used by this scoring instance, if any.

        Returns:
            associated vector model
        """

        return self.model

    def issparse(self):
        """
        Check if this scoring instance has an associated sparse keyword or sparse vector index.

        Returns:
            True if this index has an associated sparse index
        """

        raise NotImplementedError

    def isweighted(self):
        """
        Check if this scoring instance is for term weighting (i.e.) it has no associated sparse index.

        Returns:
            True if this index is for term weighting
        """

        return not self.issparse()

    def isnormalized(self):
        """
        Check if this scoring instance returns normalized scores.

        Returns:
            True if normalize is enabled, False otherwise
        """

        raise NotImplementedError



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from huggingface_hub.errors import HFValidationError
from transformers.utils.hub import cached_file

from .base import Pooling
from .cls import ClsPooling
from .mean import MeanPooling


class PoolingFactory:
    """
    Method to create pooling models.
    """

    @staticmethod
    def create(config):
        """
        Create a Pooling model.

        Args:
            config: pooling configuration

        Returns:
            Pooling
        """

        method, path, device, tokenizer, maxlength, modelargs = [
            config.get(x) for x in ["method", "path", "device", "tokenizer", "maxlength", "modelargs"]
        ]

        maxlength = PoolingFactory.maxlength(path) if isinstance(maxlength, bool) and maxlength else maxlength

        if isinstance(path, bytes) or (isinstance(path, str) and os.path.isfile(path)) or method == "pooling":
            return Pooling(path, device, tokenizer, maxlength, modelargs)

        if (not method or method not in ("clspooling", "meanpooling")) and (
            isinstance(path, str) and not os.path.exists(path)
        ):
            method = PoolingFactory.method(path)

        if method == "clspooling":
            return ClsPooling(path, device, tokenizer, maxlength, modelargs)

        return MeanPooling(path, device, tokenizer, maxlength, modelargs)

    @staticmethod
    def method(path):
        """
        Determines the pooling method using the sentence transformers pooling config.

        Args:
            path: model path

        Returns:
            pooling method
        """

        method = "meanpooling"

        config = PoolingFactory.load(path, "1_Pooling/config.json")

        if config and config["pooling_mode_cls_token"] and not config["pooling_mode_mean_tokens"]:
            method = "clspooling"

        return method

    @staticmethod
    def maxlength(path):
        """
        Reads the max_seq_length parameter from sentence transformers config.

        Args:
            path: model path

        Returns:
            max sequence length
        """

        maxlength = None

        config = PoolingFactory.load(path, "sentence_bert_config.json")
        maxlength = config.get("max_seq_length") if config else maxlength

        return maxlength

    @staticmethod
    def load(path, name):
        """
        Loads a JSON config file from the Hugging Face Hub.

        Args:
            path: model path
            name: file to load

        Returns:
            config
        """

        config = None
        try:
            path = cached_file(path_or_repo_id=path, filename=name)
            if path:
                with open(path, encoding="utf-8") as f:
                    config = json.load(f)

        except (HFValidationError, OSError):
            pass

        return config



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    import requests
    import xmltodict

    XML_TO_DICT = True
except ImportError:
    XML_TO_DICT = False

from .base import Task


class ServiceTask(Task):
    """
    Task to runs requests against remote service urls.
    """

    def register(self, url=None, method=None, params=None, batch=True, extract=None):
        """
        Adds service parameters to task. Checks if required dependencies are installed.

        Args:
            url: url to connect to
            method: http method, GET or POST
            params: default query parameters
            batch: if True, all elements are passed in a single batch request,
                        otherwise a service call is executed per element
            extract: list of sections to extract from response
        """

        if not XML_TO_DICT:
            raise ImportError('ServiceTask is not available - install "workflow" extra to enable')

        # pylint: disable=W0201
        self.url = url
        self.method = method
        self.params = params

        self.batch = batch

        self.extract = extract
        if self.extract:
            self.extract = [self.extract] if isinstance(self.extract, str) else self.extract

    def execute(self, elements, executor=None):
        if self.batch:
            elements = self.request(elements)
        else:
            elements = [self.request(element) for element in elements]

        return super().execute(elements, executor)

    def request(self, data):
        """
        Execute service request.

        Args:
            url: service url
            method: method (get or post)
            params: dict of constant parameters to pass to request
            data: dynamic data for this specific request

        Returns:
            response as JSON
        """

        if not self.params:
            params = data
        else:
            params = self.params.copy()

            for key in params:
                if not params[key]:
                    params[key] = data

        if self.method and self.method.lower() == "get":
            response = requests.get(self.url, params=params)
        else:
            response = requests.post(self.url, json=params)

        mimetype = response.headers["Content-Type"].split(";")[0]
        if mimetype.lower().endswith("xml"):
            data = xmltodict.parse(response.text)
        else:
            data = response.json()

        if self.extract:
            for tag in self.extract:
                data = data[tag]

        return data



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os


class Compress:
    """
    Base class for Compress instances.
    """

    def pack(self, path, output):
        """
        Compresses files in directory path to file output.

        Args:
            path: input directory path
            output: output file
        """

        raise NotImplementedError

    def unpack(self, path, output):
        """
        Extracts all files in path to output.

        Args:
            path: input file path
            output: output directory
        """

        raise NotImplementedError

    def validate(self, directory, path):
        """
        Validates path is under directory.

        Args:
            directory: base directory
            path: path to validate

        Returns:
            True if path is under directory, False otherwise
        """

        directory = os.path.abspath(directory)
        path = os.path.abspath(path)
        prefix = os.path.commonprefix([directory, path])

        return prefix == directory



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import operator

from .base import SQL


class Aggregate(SQL):
    """
    Aggregates partial results from queries. Partial results come from queries when working with sharded indexes.
    """

    def __init__(self, database=None):
        super().__init__(database, True)

    def __call__(self, query, results):
        """
        Analyzes query results, combines aggregate function results and applies ordering.

        Args:
            query: input query
            results: query results

        Returns:
            aggregated query results
        """

        query = super().__call__(query)

        if "select" in query:
            columns = list(results[0].keys())
            aggcolumns = self.aggcolumns(columns)
            if aggcolumns or query["orderby"]:
                if aggcolumns:
                    results = self.aggregate(query, results, columns, aggcolumns)

                return self.orderby(query, results) if query["orderby"] else self.defaultsort(results)

        return self.defaultsort(results)

    def aggcolumns(self, columns):
        """
        Filters columns for columns that have an aggregate function call.

        Args:
            columns: list of columns

        Returns:
            list of aggregate columns
        """

        aggregates = {}
        for column in columns:
            column = column.lower()
            if column.startswith(("count(", "sum(", "total(")):
                aggregates[column] = sum
            elif column.startswith("max("):
                aggregates[column] = max
            elif column.startswith("min("):
                aggregates[column] = min
            elif column.startswith("avg("):
                aggregates[column] = lambda x: sum(x) / len(x)

        return aggregates

    def aggregate(self, query, results, columns, aggcolumns):
        """
        Merges aggregate columns in results.

        Args:
            query: input query
            results: query results
            columns: list of select columns
            aggcolumns: list of aggregate columns

        Returns:
            results with aggregates merged
        """

        if query["groupby"]:
            results = self.groupby(query, results, columns)
        else:
            results = [results]

        rows = []
        for result in results:
            row = {}
            for column in columns:
                if column in aggcolumns:
                    function = aggcolumns[column]
                    row[column] = function([r[column] for r in result])
                else:
                    row[column] = result[0][column]

            rows.append(row)

        return rows

    def groupby(self, query, results, columns):
        """
        Groups results using query group by clause.

        Args:
            query: input query
            results: query results
            columns: list of select columns

        Returns:
            results grouped using group by clause
        """

        groupby = [column for column in columns if column.lower() in query["groupby"]]
        if groupby:
            results = sorted(results, key=operator.itemgetter(*groupby))
            return [list(value) for _, value in itertools.groupby(results, operator.itemgetter(*groupby))]

        return [results]

    def orderby(self, query, results):
        """
        Applies an order by clause to results.

        Args:
            query: input query
            results: query results

        Returns:
            results ordered using order by clause
        """

        for clause in query["orderby"][::-1]:
            reverse = False
            if clause.lower().endswith(" asc"):
                clause = clause.rsplit(" ")[0]
            elif clause.lower().endswith(" desc"):
                clause = clause.rsplit(" ")[0]
                reverse = True

            if clause in query["select"]:
                results = sorted(results, key=operator.itemgetter(clause), reverse=reverse)

        return results

    def defaultsort(self, results):
        """
        Default sorting algorithm for results. Sorts by score descending, if available.

        Args:
            results: query results

        Returns:
            results ordered by score descending
        """

        if results and "score" in results[0]:
            return sorted(results, key=lambda x: x["score"], reverse=True)

        return results



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import pickle
import warnings

from .base import Serialize

logger = logging.getLogger(__name__)


class Pickle(Serialize):
    """
    Pickle serialization.
    """

    def __init__(self, allowpickle=False):
        """
        Creates a new instance for Pickle serialization.

        This class ensures the allowpickle parameter or the `ALLOW_PICKLE` environment variable is True. All methods will
        raise errors if this isn't the case.

        Pickle serialization is OK for local data but it isn't recommended when sharing data externally.

        Args:
            allowpickle: default pickle allow mode, only True with methods that generate local temporary data
        """

        super().__init__()

        self.allowpickle = allowpickle

        self.version = 4

    def load(self, path):
        return super().load(path) if self.allow(path) else None

    def save(self, data, path):
        if self.allow():
            super().save(data, path)

    def loadstream(self, stream):
        return pickle.load(stream) if self.allow() else None

    def savestream(self, data, stream):
        if self.allow():
            pickle.dump(data, stream, protocol=self.version)

    def loadbytes(self, data):
        return pickle.loads(data) if self.allow() else None

    def savebytes(self, data):
        return pickle.dumps(data, protocol=self.version) if self.allow() else None

    def allow(self, path=None):
        """
        Checks if loading and saving pickled data is allowed. Raises an error if it's not allowed.

        Args:
            path: optional path to add to generated error messages
        """

        enablepickle = self.allowpickle or os.environ.get("ALLOW_PICKLE", "True") in ("True", "1")
        if not enablepickle:
            raise ValueError(
                "Loading of pickled index data is disabled. "
                f"`{path if path else 'stream'}` was not loaded. "
                "Set the env variable `ALLOW_PICKLE=True` to enable loading pickled index data. "
                "This should only be done for trusted and/or local data.",
                stacklevel=1,
            )

        if not self.allowpickle:
            warnings.warn(
                (
                    "Pickled index data formats are deprecated and loading will be disabled by default in the future. "
                    "Set the env variable `ALLOW_PICKLE=False` to disable the loading of pickled index data formats. "
                    "Saving this index will replace pickled index data formats with the latest index "
                    "formats and remove this warning.",
                ),
                FutureWarning,
                stacklevel=1,
            )

        return enablepickle



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os

from ...serialization import SerializeFactory


class Configuration:
    """
    Loads and saves index configuration.
    """

    def load(self, path):
        """
        Loads index configuration. This method supports both config.json and config pickle files.

        Args:
            path: path to directory

        Returns:
            dict
        """

        config = None

        jsonconfig = os.path.exists(f"{path}/config.json")

        name = "config.json" if jsonconfig else "config"

        with open(f"{path}/{name}", "r" if jsonconfig else "rb", encoding="utf-8" if jsonconfig else None) as handle:
            config = json.load(handle) if jsonconfig else SerializeFactory.create("pickle").loadstream(handle)

        config["format"] = "json" if jsonconfig else "pickle"

        return config

    def save(self, config, path):
        """
        Saves index configuration. This method defaults to JSON and falls back to pickle.

        Args:
            config: configuration to save
            path: path to directory

        Returns:
            dict
        """

        jsonconfig = config.get("format", "json") == "json"

        name = "config.json" if jsonconfig else "config"

        with open(f"{path}/{name}", "w" if jsonconfig else "wb", encoding="utf-8" if jsonconfig else None) as handle:
            if jsonconfig:
                json.dump(config, handle, default=str, indent=2)
            else:
                SerializeFactory.create("pickle").savestream(config, handle)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Explain class for analyzing the importance of each token in input texts for a query.
It supports permutation-based token masking to determine token importance relative to a query.
"""

import numpy as np


class Explain:
    """
    Analyzes the importance of each token in input texts for a query using permutation-based masking.

    Args:
        embeddings (Embeddings): Embeddings instance to use for similarity and scoring.
    """

    def __init__(self, embeddings):
        """
        Creates a new explain action.

        Args:
            embeddings: embeddings instance
        """

        self.embeddings = embeddings
        self.content = embeddings.config.get("content")

        self.database = embeddings.database

    def __call__(self, queries, texts, limit):
        """
        Explains the importance of each input token in text for a list of queries.

        Args:
            query: input queries
            texts: optional list of (text|list of tokens), otherwise runs search queries
            limit: optional limit if texts is None

        Returns:
            list of dict per input text per query where a higher token scores
            represents higher importance relative to the query
        """

        texts = self.texts(queries, texts, limit)

        return [self.explain(query, texts[x]) for x, query in enumerate(queries)]

    def texts(self, queries, texts, limit):
        """
        Constructs lists of dict for each input query.

        Args:
            queries: input queries
            texts: optional list of texts
            limit: optional limit if texts is None

        Returns:
            lists of dict for each input query
        """

        if texts:
            results = []
            for scores in self.embeddings.batchsimilarity(queries, texts):
                results.append([{"id": uid, "text": texts[uid], "score": score} for uid, score in scores])

            return results

        return self.embeddings.batchsearch(queries, limit) if self.content else [[]] * len(queries)

    def explain(self, query, texts):
        """
        Explains the importance of each input token in text for a list of queries.

        Args:
            query: input query
            texts: list of text

        Returns:
            list of {"id": value, "text": value, "score": value, "tokens": value} covering each input text element
        """

        results = []

        if self.database:
            query = self.database.parse(query)

            query = " ".join([" ".join(clause) for clause in query["similar"]]) if "similar" in query else None

        if not query or not texts or "score" not in texts[0] or "text" not in texts[0]:
            return texts

        for result in texts:
            text = result["text"]
            tokens = text if isinstance(text, list) else text.split()

            permutations = []
            for i in range(len(tokens)):
                data = tokens.copy()
                data.pop(i)
                permutations.append([" ".join(data)])

            scores = [(i, result["score"] - np.abs(s)) for i, s in self.embeddings.similarity(query, permutations)]

            result["tokens"] = [(tokens[i], score) for i, score in sorted(scores, key=lambda x: x[0])]

            results.append(result)

        return sorted(results, key=lambda x: x["score"], reverse=True)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Pooling


class ClsPooling(Pooling):
    """
    Builds CLS pooled vectors using outputs from a transformers model.
    """

    def forward(self, **inputs):
        """
        Runs CLS pooling on token embeddings.

        Args:
            inputs: model inputs

        Returns:
            CLS pooled embeddings using output token embeddings (i.e. last hidden state)
        """

        tokens = super().forward(**inputs)

        return tokens[:, 0]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utilities import Resolver
from .networkx import NetworkX
from .rdbms import RDBMS


class GraphFactory:
    """
    Methods to create graphs.
    """

    @staticmethod
    def create(config):
        """
        Create a Graph.

        Args:
            config: graph configuration

        Returns:
            Graph
        """

        graph = None
        backend = config.get("backend", "networkx")

        if backend == "networkx":
            graph = NetworkX(config)
        elif backend == "rdbms":
            graph = RDBMS(config)
        else:
            graph = GraphFactory.resolve(backend, config)

        config["backend"] = backend

        return graph

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            Graph
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve graph backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utilities import Resolver
from .bm25 import BM25
from .pgtext import PGText
from .sif import SIF
from .sparse import Sparse
from .tfidf import TFIDF


class ScoringFactory:
    """
    Methods to create Scoring indexes.
    """

    @staticmethod
    def create(config, models=None):
        """
        Factory method to construct a Scoring instance.

        Args:
            config: scoring configuration parameters
            models: models cache

        Returns:
            Scoring
        """

        scoring = None

        if isinstance(config, str):
            config = {"method": config}

        method = config.get("method", "bm25")

        if method == "bm25":
            scoring = BM25(config)
        elif method == "pgtext":
            scoring = PGText(config)
        elif method == "sif":
            scoring = SIF(config)
        elif method == "sparse":
            scoring = Sparse(config, models)
        elif method == "tfidf":
            scoring = TFIDF(config)
        else:
            scoring = ScoringFactory.resolve(method, config)

        config["method"] = method

        return scoring

    @staticmethod
    def issparse(config):
        """
        Checks if this scoring configuration builds a sparse index.

        Args:
            config: scoring configuration

        Returns:
            True if this config is for a sparse index
        """

        indexes = ["pgtext", "sparse"]

        return config and isinstance(config, dict) and (config.get("method") in indexes or config.get("terms"))

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: index configuration parameters

        Returns:
            Scoring
        """

        try:
            return Resolver()(backend)(config)
        except Exception as e:
            raise ImportError(f"Unable to resolve scoring backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import torch
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForQuestionAnswering,
    AutoModelForSeq2SeqLM,
    AutoModelForSequenceClassification,
    AutoTokenizer,
)
from transformers.models.auto.modeling_auto import MODEL_FOR_IMAGE_TEXT_TO_TEXT_MAPPING_NAMES

from .onnx import OnnxModel


class Models:
    """
    Utility methods for working with machine learning models
    """

    @staticmethod
    def checklength(config, tokenizer):
        """
        Checks the length for a Hugging Face Transformers tokenizer using a Hugging Face Transformers config. Copies the
        max_position_embeddings parameter if the tokenizer has no max_length set. This helps with backwards compatibility
        with older tokenizers.

        Args:
            config: transformers config
            tokenizer: transformers tokenizer
        """

        if hasattr(config, "config"):
            config = config.config

        if (
            hasattr(config, "max_position_embeddings")
            and tokenizer
            and hasattr(tokenizer, "model_max_length")
            and tokenizer.model_max_length == int(1e30)
        ):
            tokenizer.model_max_length = config.max_position_embeddings

    @staticmethod
    def maxlength(config, tokenizer):
        """
        Gets the best max length to use for generate calls. This method will return config.max_length if it's set.
        Otherwise, it will return tokenizer.model_max_length.

        Args:
            config: transformers config
            tokenizer: transformers tokenizer
        """

        if hasattr(config, "config"):
            config = config.config

        keys = config.to_diff_dict()

        return (
            config.max_length
            if "max_length" in keys or not hasattr(tokenizer, "model_max_length")
            else tokenizer.model_max_length
        )

    @staticmethod
    def deviceid(gpu):
        """
        Translates input gpu argument into a device id.

        Args:
            gpu: True/False if GPU should be enabled, also supports a device id/string/instance

        Returns:
            device id
        """

        # pylint: disable=E1101
        if isinstance(gpu, torch.device):
            return gpu

        if gpu is None or not Models.hasaccelerator():
            return -1

        if isinstance(gpu, bool):
            return 0 if gpu else -1

        return int(gpu)

    @staticmethod
    def device(deviceid):
        """
        Gets a tensor device.

        Args:
            deviceid: device id

        Returns:
            tensor device
        """

        # pylint: disable=E1101
        return deviceid if isinstance(deviceid, torch.device) else torch.device(Models.reference(deviceid))

    @staticmethod
    def reference(deviceid):
        """
        Gets a tensor device reference.

        Args:
            deviceid: device id

        Returns:
            device reference
        """

        return (
            deviceid
            if isinstance(deviceid, str)
            else (
                "cpu"
                if deviceid < 0
                else f"cuda:{deviceid}"
                if torch.cuda.is_available()
                else "mps"
                if Models.hasmpsdevice()
                else Models.finddevice()
            )
        )

    @staticmethod
    def acceleratorcount():
        """
        Gets the number of accelerator devices available.

        Returns:
            number of accelerators available
        """

        return max(torch.cuda.device_count(), int(Models.hasaccelerator()))

    @staticmethod
    def hasaccelerator():
        """
        Checks if there is an accelerator device available.

        Returns:
            True if an accelerator device is available, False otherwise
        """

        return torch.cuda.is_available() or Models.hasmpsdevice() or bool(Models.finddevice())

    @staticmethod
    def hasmpsdevice():
        """
        Checks if there is a MPS device available.

        Returns:
            True if a MPS device is available, False otherwise
        """

        return os.environ.get("PYTORCH_MPS_DISABLE") != "1" and torch.backends.mps.is_available()

    @staticmethod
    def finddevice():
        """
        Attempts to find an alternative accelerator device.

        Returns:
            name of first alternative accelerator available or None if not found
        """

        return next(
            (device for device in ["xpu"] if hasattr(torch, device) and getattr(torch, device).is_available()), None
        )

    @staticmethod
    def load(path, config=None, task="default", modelargs=None):
        """
        Loads a machine learning model. Handles multiple model frameworks (ONNX, Transformers).

        Args:
            path: path to model
            config: path to model configuration
            task: task name used to lookup model type

        Returns:
            machine learning model
        """

        if isinstance(path, bytes) or (isinstance(path, str) and os.path.isfile(path)):
            return OnnxModel(path, config)

        if not isinstance(path, str):
            return path

        models = {
            "default": AutoModel.from_pretrained,
            "question-answering": AutoModelForQuestionAnswering.from_pretrained,
            "summarization": AutoModelForSeq2SeqLM.from_pretrained,
            "text-classification": AutoModelForSequenceClassification.from_pretrained,
            "zero-shot-classification": AutoModelForSequenceClassification.from_pretrained,
        }

        modelargs = modelargs if modelargs else {}

        return models[task](path, **modelargs) if task in models else path

    @staticmethod
    def tokenizer(path, **kwargs):
        """
        Loads a tokenizer from path.

        Args:
            path: path to tokenizer
            kwargs: optional additional keyword arguments

        Returns:
            tokenizer
        """

        return AutoTokenizer.from_pretrained(path, **kwargs) if isinstance(path, str) else path

    @staticmethod
    def task(path, **kwargs):
        """
        Attempts to detect the model task from path.

        Args:
            path: path to model
            kwargs: optional additional keyword arguments

        Returns:
            inferred model task
        """

        config = None
        if isinstance(path, list | tuple) and hasattr(path[0], "config"):
            config = path[0].config
        elif isinstance(path, str):
            config = AutoConfig.from_pretrained(path, **kwargs)

        task = None
        if config:
            architecture = config.architectures[0] if config.architectures else None
            if architecture:
                if architecture in MODEL_FOR_IMAGE_TEXT_TO_TEXT_MAPPING_NAMES.values():
                    task = "vision"
                elif any(x for x in ["LMHead", "CausalLM"] if x in architecture):
                    task = "language-generation"
                elif "QuestionAnswering" in architecture:
                    task = "question-answering"
                elif "ConditionalGeneration" in architecture:
                    task = "sequence-sequence"

        return task



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter

from .topics import Topics


# pylint: disable=R0904
class Graph:
    """
    Base class for Graph instances. This class builds graph networks. Supports topic modeling
    and relationship traversal.
    """

    def __init__(self, config):
        """
        Creates a new Graph.

        Args:
            config: graph configuration
        """

        self.config = config if config is not None else {}

        self.backend = None

        self.categories = None
        self.topics = None

        columns = config.get("columns", {})
        self.text = columns.get("text", "text")
        self.object = columns.get("object", "object")

        self.copyattributes = config.get("copyattributes", False)

        self.relationships = columns.get("relationships", "relationships")
        self.relations = {}

    def create(self):
        """
        Creates the graph network.
        """

        raise NotImplementedError

    def count(self):
        """
        Returns the total number of nodes in graph.

        Returns:
            total nodes in graph
        """

        raise NotImplementedError

    def scan(self, attribute=None, data=False):
        """
        Iterates over nodes that match a criteria. If no criteria specified, all nodes
        are returned.

        Args:
            attribute: if specified, nodes having this attribute are returned
            data: if True, attribute data is also returned

        Returns:
            node id iterator if data is False or (id, attribute dictionary) iterator if data is True
        """

        raise NotImplementedError

    def node(self, node):
        """
        Get node by id. Returns None if not found.

        Args:
            node: node id

        Returns:
            graph node
        """

        raise NotImplementedError

    def addnode(self, node, **attrs):
        """
        Adds a node to the graph.

        Args:
            node: node id
            attrs: node attributes
        """

        raise NotImplementedError

    def addnodes(self, nodes):
        """
        Adds nodes to the graph.

        Args:
            nodes: list of (node, attributes) to add
        """

        raise NotImplementedError

    def removenode(self, node):
        """
        Removes a node and all it's edges from graph.

        Args:
            node: node id
        """

        raise NotImplementedError

    def hasnode(self, node):
        """
        Returns True if node found, False otherwise.

        Args:
            node: node id

        Returns:
            True if node found, False otherwise
        """

        raise NotImplementedError

    def attribute(self, node, field):
        """
        Gets a node attribute.

        Args:
            node: node id
            field: attribute name

        Returns:
            attribute value
        """

        raise NotImplementedError

    def addattribute(self, node, field, value):
        """
        Adds an attribute to node.

        Args:
            node: node id
            field: attribute name
            value: attribute value
        """

        raise NotImplementedError

    def removeattribute(self, node, field):
        """
        Removes an attribute from node.

        Args:
            node: node id
            field: attribute name

        Returns:
            attribute value or None if not present
        """

        raise NotImplementedError

    def edgecount(self):
        """
        Returns the total number of edges.

        Returns:
            total number of edges in graph
        """

        raise NotImplementedError

    def edges(self, node):
        """
        Gets edges of node by id.

        Args:
            node: node id

        Returns:
            list of edge node ids
        """

        raise NotImplementedError

    def addedge(self, source, target, **attrs):
        """
        Adds an edge to graph.

        Args:
            source: node 1 id
            target: node 2 id
        """

        raise NotImplementedError

    def addedges(self, edges):
        """
        Adds an edge to graph.

        Args:
            edges: list of (source, target, attributes) to add
        """

        raise NotImplementedError

    def hasedge(self, source, target=None):
        """
        Returns True if edge found, False otherwise. If target is None, this method
        returns True if any edge is found.

        Args:
            source: node 1 id
            target: node 2 id

        Returns:
            True if edge found, False otherwise
        """

        raise NotImplementedError

    def centrality(self):
        """
        Runs a centrality algorithm on the graph.

        Returns:
            dict of {node id: centrality score}
        """

        raise NotImplementedError

    def pagerank(self):
        """
        Runs the pagerank algorithm on the graph.

        Returns:
            dict of {node id, page rank score}
        """

        raise NotImplementedError

    def showpath(self, source, target):
        """
        Gets the shortest path between source and target.

        Args:
            source: start node id
            target: end node id

        Returns:
            list of node ids representing the shortest path
        """

        raise NotImplementedError

    def isquery(self, queries):
        """
        Checks if queries are supported graph queries.

        Args:
            queries: queries to check

        Returns:
            True if all the queries are supported graph queries, False otherwise
        """

        raise NotImplementedError

    def parse(self, query):
        """
        Parses a graph query into query components.

        Args:
            query: graph query

        Returns:
            query components as a dictionary
        """

        raise NotImplementedError

    def search(self, query, limit=None, graph=False):
        """
        Searches graph for nodes matching query.

        Args:
            query: graph query
            limit: maximum results
            graph: return graph results if True

        Returns:
            list of dict if graph is set to False
            filtered graph if graph is set to True
        """

        raise NotImplementedError

    def batchsearch(self, queries, limit=None, graph=False):
        """
        Searches graph for nodes matching query.

        Args:
            query: graph query
            limit: maximum results
            graph: return graph results if True

        Returns:
            list of dict if graph is set to False
            filtered graph if graph is set to True
        """

        return [self.search(query, limit, graph) for query in queries]

    def communities(self, config):
        """
        Run community detection on the graph.

        Args:
            config: configuration

        Returns:
            dictionary of {topic name:[ids]}
        """

        raise NotImplementedError

    def load(self, path):
        """
        Loads a graph at path.

        Args:
            path: path to graph
        """

        raise NotImplementedError

    def save(self, path):
        """
        Saves a graph at path.

        Args:
            path: path to save graph
        """

        raise NotImplementedError

    def loaddict(self, data):
        """
        Loads data from input dictionary into this graph.

        Args:
            data: input dictionary
        """

        raise NotImplementedError

    def savedict(self):
        """
        Saves graph data to a dictionary.

        Returns:
            dict
        """

        raise NotImplementedError

    def initialize(self):
        """
        Initialize graph instance.
        """

        if not self.backend:
            self.backend = self.create()

    def close(self):
        """
        Closes this graph.
        """

        self.backend, self.categories, self.topics = None, None, None

    def insert(self, documents, index=0):
        """
        Insert graph nodes for each document.

        Args:
            documents: list of (id, data, tags)
            index: indexid offset, used for node ids
        """

        self.initialize()

        nodes = []
        for uid, document, _ in documents:
            relations, attributes = None, {}

            if isinstance(document, dict):
                relations = document.get(self.relationships)

                search = self.copyattributes if isinstance(self.copyattributes, list) else []
                attributes = {
                    k: v
                    for k, v in document.items()
                    if k not in [self.text, self.object, self.relationships]
                    and (self.copyattributes is True or k in search)
                }

                document = document.get(self.text, document.get(self.object))

            if document is not None:
                if isinstance(document, list):
                    document = " ".join(document)

                nodes.append((index, {**{"id": uid, "data": document}, **attributes}))

                self.addrelations(index, relations)

                index += 1

        self.addnodes(nodes)

    def delete(self, ids):
        """
        Deletes ids from graph.

        Args:
            ids: node ids to delete
        """

        for node in ids:
            if self.hasnode(node):
                topic = self.attribute(node, "topic")
                if topic and self.topics:
                    self.topics[topic].remove(node)

                    if not self.topics[topic]:
                        self.topics.pop(topic)

                self.removenode(node)

    def index(self, search, ids, similarity):
        """
        Build relationships between graph nodes using a score-based search function.

        Args:
            search: batch search function - takes a list of queries and returns lists of (id, scores)
            ids: ids function - internal id resolver
            similarity: batch similarity function - takes a list of text and labels and returns best matches
        """

        self.resolverelations(ids)

        self.inferedges(self.scan(), search)

        if "topics" in self.config:
            self.addtopics(similarity)

    def upsert(self, search, ids, similarity=None):
        """
        Adds relationships for new graph nodes using a score-based search function.

        Args:
            search: batch search function - takes a list of queries and returns lists of (id, scores)
            ids: ids function - internal id resolver
            similarity: batch similarity function - takes a list of text and labels and returns best matches
        """

        hastopics = "topics" in self.config

        self.resolverelations(ids)

        self.inferedges(self.scan(attribute="data"), search, {"updated": True} if hastopics else None)

        if hastopics:
            if self.topics:
                self.infertopics()
            else:
                self.addtopics(similarity)

    def filter(self, nodes, graph=None):
        """
        Creates a subgraph of this graph using the list of input nodes. This method creates a new graph
        selecting only matching nodes, edges, topics and categories.

        Args:
            nodes: nodes to select as a list of ids or list of (id, score) tuples
            graph: optional graph used to store filtered results

        Returns:
            graph
        """

        graph = graph if graph else type(self)(self.config)

        graph.initialize()

        nodeids = {node[0] if isinstance(node, tuple) else node for node in nodes}
        for node in nodes:
            node, score = node if isinstance(node, tuple) else (node, None)

            graph.addnode(node, **self.node(node))

            if score is not None:
                graph.addattribute(node, "score", score)

            edges = self.edges(node)
            if edges:
                for target, attributes in self.edges(node).items():
                    if target in nodeids:
                        graph.addedge(node, target, **attributes)

        if self.topics:
            topics = {}
            for i, (topic, ids) in enumerate(self.topics.items()):
                ids = [x for x in ids if x in nodeids]
                if ids:
                    topics[topic] = (self.categories[i] if self.categories else None, ids)

            topics = sorted(topics.items(), key=lambda x: len(x[1][1]), reverse=True)

            graph.categories = [category for _, (category, _) in topics] if self.categories else None
            graph.topics = {topic: ids for topic, (_, ids) in topics}

        return graph

    def addrelations(self, node, relations):
        """
        Add manually-provided relationships.

        Args:
            node: node id
            relations: list of relationships to add
        """

        if relations:
            if node not in self.relations:
                self.relations[node] = []

            for relation in relations:
                relation = {"id": relation} if not isinstance(relation, dict) else relation
                self.relations[node].append(relation)

    def resolverelations(self, ids):
        """
        Resolves ids and creates edges for manually-provided relationships.

        Args:
            ids: internal id resolver
        """

        edges = []

        for node, relations in self.relations.items():
            iids = ids(y["id"] for y in relations)

            for relation in relations:
                relation = relation.copy()

                targets = iids.get(str(relation.pop("id")))

                if targets:
                    for target in targets:
                        relation["weight"] = relation.get("weight", 1.0)

                        edges.append((node, target, relation))

        if edges:
            self.addedges(edges)

        self.relations = {}

    def inferedges(self, nodes, search, attributes=None):
        """
        Infers edges for a list of nodes using a score-based search function.

        Args:
            nodes: list of nodes
            search: search function to use to identify edges
            attribute: dictionary of attributes to add to each node
        """

        batchsize, limit, minscore = (
            self.config.get("batchsize", 256),
            self.config.get("limit", 15),
            self.config.get("minscore", 0.1),
        )
        approximate = self.config.get("approximate", True)

        batch = []
        for node in nodes:
            data = self.removeattribute(node, "data")

            if isinstance(data, str):
                self.addattribute(node, "text", data)

            if attributes:
                for field, value in attributes.items():
                    self.addattribute(node, field, value)

            if not approximate or not self.hasedge(node):
                batch.append((node, data))

            if len(batch) == batchsize:
                self.addbatch(search, batch, limit, minscore)
                batch = []

        if batch:
            self.addbatch(search, batch, limit, minscore)

    def addbatch(self, search, batch, limit, minscore):
        """
        Adds batch of documents to graph. This method runs the search function for each item in batch
        and adds node edges between the input and each search result.

        Args:
            search: search function to use to identify edges
            batch: batch to add
            limit: max edges to add per node
            minscore: min score to add node edge
        """

        edges = []
        for x, result in enumerate(search([data for _, data in batch], limit)):
            x, _ = batch[x]

            for y, score in result:
                if str(x) != str(y) and score > minscore:
                    edges.append((x, y, {"weight": score}))

        self.addedges(edges)

    def addtopics(self, similarity=None):
        """
        Identifies and adds topics using community detection.

        Args:
            similarity: similarity function for labeling categories
        """

        self.cleartopics()

        topics = Topics(self.config["topics"])
        config = topics.config
        self.topics = topics(self)

        if "categories" in config and similarity:
            self.categories = []
            results = similarity(self.topics.keys(), config["categories"])
            for result in results:
                self.categories.append(config["categories"][result[0][0]])

        for x, topic in enumerate(self.topics):
            for r, node in enumerate(self.topics[topic]):
                self.addattribute(node, "topic", topic)
                self.addattribute(node, "topicrank", r)

                if self.categories:
                    self.addattribute(node, "category", self.categories[x])

    def cleartopics(self):
        """
        Clears topic fields from all nodes.
        """

        if self.topics:
            for node in self.scan():
                self.removeattribute(node, "topic")
                self.removeattribute(node, "topicrank")

                if self.categories:
                    self.removeattribute(node, "category")

            self.topics, self.categories = None, None

    def infertopics(self):
        """
        Infers topics for all nodes with an "updated" attribute. This method analyzes the direct node
        neighbors and set the most commonly occuring topic and category for each node.
        """

        for node in self.scan(attribute="updated"):
            self.removeattribute(node, "updated")

            ids = self.edges(node)
            ids = ids.keys() if ids else None

            topic = Counter(self.attribute(x, "topic") for x in ids).most_common(1)[0][0] if ids else None
            if topic:
                self.topics[topic].append(node)
                self.addattribute(node, "topic", topic)

                self.addattribute(node, "topicrank", len(self.topics[topic]) - 1)

                category = Counter(self.attribute(x, "category") for x in ids).most_common(1)[0][0]
                self.addattribute(node, "category", category)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    import litellm as api

    LITELLM = True
except ImportError:
    LITELLM = False

from ..base import Vectors


class LiteLLM(Vectors):
    """
    Builds vectors using an external embeddings API via LiteLLM.
    """

    @staticmethod
    def ismodel(path):
        """
        Checks if path is a LiteLLM model.

        Args:
            path: input path

        Returns:
            True if this is a LiteLLM model, False otherwise
        """

        # pylint: disable=W0702
        if isinstance(path, str) and LITELLM:
            debug = api.suppress_debug_info
            try:
                api.suppress_debug_info = True
                return api.get_llm_provider(path)
            except Exception:
                return False
            finally:
                api.suppress_debug_info = debug

        return False

    def __init__(self, config, scoring, models):
        if not LITELLM:
            raise ImportError('LiteLLM is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        return None

    def encode(self, data, category=None):
        response = api.embedding(model=self.config.get("path"), input=data, **self.config.get("vectors", {}))

        return np.array([x["embedding"] for x in response.data], dtype=np.float32)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile

from ..compression import ArchiveFactory
from ..databases import DatabaseFactory
from ..graph_algorithms import GraphFactory
from ..remote_storage import CloudFactory
from ..score_functions import ScoringFactory
from ..vector_index import ANNFactory
from ..vector_processing import VectorsFactory
from .embedding_index import Action, Configuration, Functions, Indexes, IndexIds, Reducer, Stream, Transform
from .embedding_search import Explain, Ids, Query, Search, Terms
from .embedding_search.base import SearchResult

# pylint: disable=C0302,R0904
# pyright: reportOptionalSubscript=none
# pyright: reportOptionalOperand=none

"""
This module provides the Embeddings class, which manages semantic search and vector indexing for text and other data.
It supports building, updating, searching, and maintaining embedding indexes using various backends and algorithms.
"""


class Embeddings:
    """
    Manages embeddings databases for semantic search and vector indexing.

    The Embeddings class transforms data into embedding vectors, builds and maintains indexes, and enables semantic search
    where similar concepts yield similar vectors. It supports dense and sparse vector models, multiple backends, and
    advanced features like upsert, reindex, and graph-based search.

    Attributes:
        config (dict): Embeddings configuration.
        models (Any): Optional models cache for sharing between embeddings.
        reducer, model, ann, ids, database, functions, graph, scoring, query, archive, indexes: Internal components for processing and indexing.
    """

    # pylint: disable = W0231
    def __init__(self, config=None, models=None, **kwargs):
        """
        Initializes a new Embeddings index.

        Embeddings indexes are thread-safe for read operations, but writes must be synchronized externally.

        Args:
            config (dict, optional): Embeddings configuration dictionary.
            models (Any, optional): Models cache for sharing between embeddings instances.
            **kwargs: Additional configuration parameters.
        """

        self.config: dict | None = None
        self.reducer = None
        self.model = None
        self.ann = None
        self.ids = None
        self.database = None
        self.functions = None
        self.graph = None
        self.scoring = None
        self.query = None
        self.archive = None
        self.indexes = None
        self.models = models
        config = {**config, **kwargs} if config and kwargs else kwargs if kwargs else config
        self.configure(config)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def score(self, documents):
        """
        Builds a term weighting scoring index. Only used by word vector models.

        Args:
            documents (Iterable): Iterable of (id, data, tags), (id, data), or data.
        """

        if self.isweighted():
            self.scoring.index(Stream(self)(documents))

    def index(self, documents, reindex=False, checkpoint=None):
        """
        Builds an embeddings index, overwriting any existing index.

        Args:
            documents (Iterable): Iterable of (id, data, tags), (id, data), or data.
            reindex (bool, optional): If True, performs a reindex operation and skips database creation. Defaults to False.
            checkpoint (str, optional): Optional checkpoint directory to enable indexing restart.
        """

        self.initindex(reindex)

        transform = Transform(self, Action.REINDEX if reindex else Action.INDEX, checkpoint)
        stream = Stream(self, Action.REINDEX if reindex else Action.INDEX)

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".npy") as buffer:
            ids, dimensions, embeddings = transform(stream(documents), buffer)
            if embeddings is not None:
                if self.config.get("pca"):
                    self.reducer = Reducer(embeddings, self.config["pca"])
                    self.reducer(embeddings)

                self.config["dimensions"] = dimensions

                self.ann = self.createann()

                self.ann.index(embeddings)

            if ids and not reindex and not self.database:
                self.ids = self.createids(ids)

        if self.issparse():
            self.scoring.index()

        if self.indexes:
            self.indexes.index()

        if self.graph:
            self.graph.index(Search(self, indexonly=True), Ids(self), self.batchsimilarity)

    def upsert(self, documents, checkpoint=None):
        """
        Runs an embeddings upsert operation. Appends new data or updates existing data in the index.
        If the index does not exist, performs a standard index operation.

        Args:
            documents (Iterable): Iterable of (id, data, tags), (id, data), or data.
            checkpoint (str, optional): Optional checkpoint directory to enable indexing restart.
        """

        if not self.count():
            self.index(documents, checkpoint=checkpoint)
            return

        transform = Transform(self, Action.UPSERT, checkpoint=checkpoint)
        stream = Stream(self, Action.UPSERT)

        with tempfile.NamedTemporaryFile(mode="wb", suffix=".npy") as buffer:
            ids, _, embeddings = transform(stream(documents), buffer)
            if embeddings is not None:
                if self.reducer:
                    self.reducer(embeddings)

                self.ann.append(embeddings)

            if ids and not self.database:
                self.ids = self.createids(self.ids + ids)

        if self.issparse():
            self.scoring.upsert()

        if self.indexes:
            self.indexes.upsert()

        if self.graph:
            self.graph.upsert(Search(self, indexonly=True), Ids(self), self.batchsimilarity)

    def delete(self, ids):
        """
        Deletes entries from the embeddings index.

        Args:
            ids (List): List of ids to delete.

        Returns:
            List: List of ids that were deleted.
        """

        indices = []

        deletes = []

        if self.database:
            ids = self.database.ids(ids)

            indices = [i for i, _ in ids]
            deletes = sorted(set(uid for _, uid in ids))

            self.database.delete(deletes)
        elif self.ann or self.scoring:
            for uid in ids:
                indices.extend([index for index, value in enumerate(self.ids) if uid == value])

            for index in indices:
                deletes.append(self.ids[index])
                self.ids[index] = None

        if indices:
            if self.isdense():
                self.ann.delete(indices)

            if self.issparse():
                self.scoring.delete(indices)

            if self.indexes:
                self.indexes.delete(indices)

            if self.graph:
                self.graph.delete(indices)

        return deletes

    def reindex(self, config=None, function=None, **kwargs):
        """
        Recreates the embeddings index using a new configuration.
        Only works if document content storage is enabled.

        Args:
            config (dict, optional): New configuration for the embeddings index.
            function (Callable, optional): Optional function to prepare content for indexing.
            **kwargs: Additional configuration parameters.
        """

        if self.database:
            config = {**config, **kwargs} if config and kwargs else config if config else kwargs

            config["content"] = self.config["content"]
            if "objects" in self.config:
                config["objects"] = self.config["objects"]

            self.configure(config)

            if self.functions:
                self.functions.reset()

            if function:
                self.index(function(self.database.reindex(self.config)), True)
            else:
                self.index(self.database.reindex(self.config), True)

    def transform(self, document, category=None, index=None):
        """
        Transforms a document into an embeddings vector.

        Args:
            document (Any): Single document to transform (id, data, tags), (id, data), or data.
            category (str, optional): Category for instruction-based embeddings.
            index (str, optional): Index name, if applicable.

        Returns:
            Any: Embeddings vector for the document.
        """

        return self.batchtransform([document], category, index)[0]

    def batchtransform(self, documents, category=None, index=None):
        """
        Transforms multiple documents into embeddings vectors.

        Args:
            documents (Iterable): Iterable of (id, data, tags), (id, data), or data.
            category (str, optional): Category for instruction-based embeddings.
            index (str, optional): Index name, if applicable.

        Returns:
            Any: Embeddings vectors for the documents.
        """

        self.defaults()

        model = self.findmodel(index)

        embeddings = model.batchtransform(Stream(self)(documents), category)

        if self.reducer:
            self.reducer(embeddings)

        return embeddings

    def count(self):
        """
        Returns the total number of elements in this embeddings index.

        Returns:
            int: Number of elements in this embeddings index.
        """

        if self.ann:
            return self.ann.count()
        if self.scoring:
            return self.scoring.count()
        if self.database:
            return self.database.count()
        if self.ids:
            return len([uid for uid in self.ids if uid is not None])

        return 0

    def search(self, query, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input query.
        Runs an index search, index + database search, or a graph search depending on configuration and query.

        Args:
            query (Any): Input query.
            limit (int, optional): Maximum number of results.
            weights (Any, optional): Hybrid score weights, if applicable.
            index (str, optional): Index name, if applicable.
            parameters (dict, optional): Named parameters to bind to placeholders.
            graph (bool, optional): If True, return graph results.

        Returns:
            list: List of (id, score) for index search, list of dict for index + database search, or graph if graph is True.
        """

        results = self.batchsearch([query], limit, weights, index, [parameters], graph)
        return results[0] if results else results

    def batchsearch(self, queries, limit=None, weights=None, index=None, parameters=None, graph=False):
        """
        Finds documents most similar to the input queries.
        Runs an index search, index + database search, or a graph search depending on configuration and query.

        Args:
            queries (Iterable): Input queries.
            limit (int, optional): Maximum number of results per query.
            weights (Any, optional): Hybrid score weights, if applicable.
            index (str, optional): Index name, if applicable.
            parameters (list, optional): List of dicts of named parameters to bind to placeholders.
            graph (bool, optional): If True, return graph results.

        Returns:
            list: List of (id, score) per query for index search, list of dict per query for index + database search, or list of graphs if graph is True.
        """

        graph = graph if self.graph else False

        results = Search(self, indexids=graph)(queries, limit, weights, index, parameters)

        return [self.graph.filter(x) if isinstance(x, list) else x for x in results] if graph else results

    def similarity(self, query, data):
        """
        Computes the similarity between query and list of data. Returns a list of
        SearchResult(index, score) sorted by highest score, where index is the index in data.

        Args:
            query: input query
            data: list of data

        Returns:
            list of SearchResult(index, score)
        """

        return self.batchsimilarity([query], data)[0]

    def batchsimilarity(self, queries, data):
        """
        Computes the similarity between list of queries and list of data. Returns a list
        of SearchResult(index, score) sorted by highest score per query, where index is the index in data.

        Args:
            queries: input queries
            data: list of data

        Returns:
            list of SearchResult(index, score) per query
        """

        queries = self.batchtransform(((None, query, None) for query in queries), "query")
        data = self.batchtransform(((None, row, None) for row in data), "data")

        model = self.findmodel()

        scores = model.dot(queries, data)

        return [
            [SearchResult(index=i, score=s) for i, s in sorted(enumerate(score), key=lambda x: x[1], reverse=True)]
            for score in scores
        ]

    def explain(self, query, texts=None, limit=None):
        """
        Explains the importance of each input token in text for a query. This method requires
        either content to be enabled or texts to be provided.

        Args:
            query: input query
            texts: optional list of (text|list of tokens), otherwise runs search query
            limit: optional limit if texts is None

        Returns:
            list of dict per input text where a higher token scores represents higher importance relative to the query
        """

        results = self.batchexplain([query], texts, limit)
        return results[0] if results else results

    def batchexplain(self, queries, texts=None, limit=None):
        """
        Explains the importance of each input token in text for a list of queries.
        This method requires either content to be enabled or texts to be provided.

        Args:
            queries: input queries
            texts: optional list of (text|list of tokens), otherwise runs search queries
            limit: optional limit if texts is None

        Returns:
            list of dict per input text per query where a higher token scores represents
            higher importance relative to the query
        """

        return Explain(self)(queries, texts, limit)

    def terms(self, query):
        """
        Extracts keyword terms from a query.

        Args:
            query: input query

        Returns:
            query reduced down to keyword terms
        """

        return self.batchterms([query])[0]

    def batchterms(self, queries):
        """
        Extracts keyword terms from a list of queries.

        Args:
            queries: list of queries

        Returns:
            list of queries reduced down to keyword term strings
        """

        return Terms(self)(queries)

    def exists(self, path=None, cloud=None, **kwargs):
        """
        Checks if an index exists at path.

        Args:
            path: input path
            cloud: cloud storage configuration
            kwargs: additional configuration as keyword args

        Returns:
            True if index exists, False otherwise
        """

        cloud = self.createcloud(cloud=cloud, **kwargs)
        if cloud:
            return cloud.exists(path)

        path, apath = self.checkarchive(path)
        if apath:
            return os.path.exists(apath)

        return (
            path
            and (os.path.exists(f"{path}/config.json") or os.path.exists(f"{path}/config"))
            and "offset" in Configuration().load(path)
        )

    def load(self, path=None, cloud=None, config=None, **kwargs):
        """
        Loads an existing index from path.

        Args:
            path: input path
            cloud: cloud storage configuration
            config: configuration overrides
            kwargs: additional configuration as keyword args

        Returns:
            Embeddings
        """

        cloud = self.createcloud(cloud=cloud, **kwargs)
        if cloud:
            path = cloud.load(path)

        path, apath = self.checkarchive(path)
        if apath:
            self.archive.load(apath)

        self.config = Configuration().load(path)

        self.config = {**self.config, **config} if config else self.config

        self.ann = self.createann()
        if self.ann:
            self.ann.load(f"{path}/embeddings")

        if self.config.get("pca"):
            self.reducer = Reducer()
            self.reducer.load(f"{path}/lsa")

        self.ids = self.createids()
        if self.ids:
            self.ids.load(f"{path}/ids")

        self.database = self.createdatabase()
        if self.database:
            self.database.load(f"{path}/documents")

        self.scoring = self.createscoring()
        if self.scoring:
            self.scoring.load(f"{path}/scoring")

        self.indexes = self.createindexes()
        if self.indexes:
            self.indexes.load(f"{path}/indexes")

        self.graph = self.creategraph()
        if self.graph:
            self.graph.load(f"{path}/graph")

        self.model = self.loadvectors()

        self.query = self.loadquery()

        return self

    def save(self, path, cloud=None, **kwargs):
        """
        Saves an index in a directory at path unless path ends with tar.gz, tar.bz2, tar.xz or zip.
        In those cases, the index is stored as a compressed file.

        Args:
            path: output path
            cloud: cloud storage configuration
            kwargs: additional configuration as keyword args
        """

        if self.config:
            path, apath = self.checkarchive(path)
            os.makedirs(path, exist_ok=True)
            Configuration().save(self.config, path)
            if self.ann:
                self.ann.save(f"{path}/embeddings")
            if self.reducer:
                self.reducer.save(f"{path}/lsa")
            if self.ids:
                self.ids.save(f"{path}/ids")
            if self.database:
                self.database.save(f"{path}/documents")
            if self.scoring:
                self.scoring.save(f"{path}/scoring")
            if self.indexes:
                self.indexes.save(f"{path}/indexes")
            if self.graph:
                self.graph.save(f"{path}/graph")
            if apath:
                self.archive.save(apath)
            cloud = self.createcloud(cloud=cloud, **kwargs)
            if cloud:
                cloud.save(apath if apath else path)

    def close(self):
        """
        Closes this embeddings index and frees all resources.
        """
        self.config, self.archive = None, None
        self.reducer, self.query = None, None
        self.ids = None
        if self.ann:
            self.ann.close()
            self.ann = None
        if self.database:
            self.database.close()
            self.database, self.functions = None, None
        if self.scoring:
            self.scoring.close()
            self.scoring = None
        if self.graph:
            self.graph.close()
            self.graph = None
        if self.indexes:
            self.indexes.close()
            self.indexes = None
        if self.model:
            self.model.close()
            self.model = None
        self.models = None

    def info(self):
        """
        Prints the current embeddings index configuration.
        """

        if self.config:
            print(json.dumps(self.config, sort_keys=True, default=str, indent=2))

    def issparse(self):
        """
        Checks if this instance has an associated sparse keyword or sparse vectors scoring index.

        Returns:
            True if scoring has an associated sparse keyword/vector index, False otherwise
        """

        return self.scoring and self.scoring.issparse()

    def isdense(self):
        """
        Checks if this instance has an associated VectoreIndex instance.

        Returns:
            True if this instance has an associated VectoreIndex, False otherwise
        """

        return self.ann is not None

    def isweighted(self):
        """
        Checks if this instance has an associated scoring instance with term weighting enabled.

        Returns:
            True if term weighting is enabled, False otherwise
        """

        return self.scoring and self.scoring.isweighted()

    def findmodel(self, index=None):
        """
        Finds the primary vector model used by this instance.

        Returns:
            Vectors
        """

        return (
            self.indexes.findmodel(index)
            if index and self.indexes
            else (
                self.model
                if self.model
                else self.scoring.findmodel()
                if self.scoring and self.scoring.findmodel()
                else self.indexes.findmodel()
                if self.indexes
                else None
            )
        )

    def configure(self, config):
        """
        Sets the configuration for this embeddings index and loads config-driven models.

        Args:
            config: embeddings configuration
        """

        self.config = config

        self.reducer = None

        scoring = self.config.get("scoring") if self.config else None
        self.scoring = self.createscoring() if scoring and not self.hassparse() else None

        self.model = self.loadvectors() if self.config else None

        self.query = self.loadquery() if self.config else None

    def initindex(self, reindex):
        """
        Initialize new index.

        Args:
            reindex: if this is a reindex operation in which case database creation is skipped, defaults to False
        """

        self.defaults()

        self.ids = None

        if not reindex:
            self.database = self.createdatabase()

            self.archive = None

        if self.ann:
            self.ann.close()

        self.ann = None

        if self.hassparse():
            self.scoring = self.createscoring()

        self.indexes = self.createindexes()

        self.graph = self.creategraph()

    def defaults(self):
        """
        Apply default parameters to current configuration.

        Returns:
            configuration with default parameters set
        """

        self.config = self.config if self.config else {}

        if not self.config.get("scoring") and any(self.config.get(key) for key in ["keyword", "sparse", "hybrid"]):
            self.defaultsparse()

        if self.config.get("graph") is True:
            self.config["graph"] = {}

        if not self.model and (self.defaultallowed() or self.config.get("dense")):
            self.config["path"] = "sentence-transformers/all-MiniLM-L6-v2"

            self.model = self.loadvectors()

    def defaultsparse(self):
        """
        Logic to derive default sparse index configuration.
        """

        method = None
        for x in ["keyword", "hybrid"]:
            value = self.config.get(x)
            if value:
                method = value if isinstance(value, str) else "bm25"

                if x == "hybrid":
                    self.config["dense"] = True

        sparse = self.config.get("sparse", {})
        if sparse or method == "sparse":
            sparse = (
                {"path": self.config.get("sparse")}
                if isinstance(sparse, str)
                else {}
                if isinstance(sparse, bool)
                else sparse
            )
            sparse["path"] = sparse.get("path", "prithivida/Splade_PP_en_v2")

            self.config["scoring"] = {**{"method": "sparse"}, **sparse}

        elif method:
            self.config["scoring"] = {"method": method, "terms": True, "normalize": True}

    def defaultallowed(self):
        """
        Tests if this embeddings instance can use a default model if not otherwise provided.

        Returns:
            True if a default model is allowed, False otherwise
        """

        params = [("keyword", False), ("sparse", False), ("defaults", True)]
        return all(self.config.get(key, default) == default for key, default in params)

    def loadvectors(self):
        """
        Loads a vector model set in config.

        Returns:
            vector model
        """

        if "indexes" in self.config and self.models is None:
            self.models = {}

        dense = self.config.get("dense")
        if not self.config.get("path") and dense and isinstance(dense, str):
            self.config["path"] = dense

        return VectorsFactory.create(self.config, self.scoring, self.models)

    def loadquery(self):
        """
        Loads a query model set in config.

        Returns:
            query model
        """

        if "query" in self.config:
            return Query(**self.config["query"])

        return None

    def checkarchive(self, path):
        """
        Checks if path is an archive file.

        Args:
            path: path to check

        Returns:
            (working directory, current path) if this is an archive, original path otherwise
        """
        self.archive = ArchiveFactory.create()
        if self.archive.isarchive(path):
            return self.archive.path(), path
        return path, None

    def createcloud(self, **cloud):
        """
        Creates a cloud instance from config.

        Args:
            cloud: cloud configuration
        """

        config = cloud
        if config.get("cloud"):
            config.update(config.pop("cloud"))

        return CloudFactory.create(config) if config else None

    def createann(self):
        """
        Creates an VectoreIndex from config.

        Returns:
            new VectoreIndex, if enabled in config
        """

        if self.ann:
            self.ann.close()

        return ANNFactory.create(self.config) if self.config.get("path") or self.defaultallowed() else None

    def createdatabase(self):
        """
        Creates a database from config. This method will also close any existing database connection.

        Returns:
            new database, if enabled in config
        """

        if self.database:
            self.database.close()

        config = self.config.copy()

        self.functions = Functions(self) if "functions" in config else None
        if self.functions:
            config["functions"] = self.functions(config)

        return DatabaseFactory.create(config)

    def creategraph(self):
        """
        Creates a graph from config.

        Returns:
            new graph, if enabled in config
        """

        if self.graph:
            self.graph.close()

        if "graph" in self.config:
            config = self.config["graph"] if "graph" in self.config else {}

            config = self.columns(config)
            return GraphFactory.create(config)

        return None

    def createids(self, ids=None):
        """
        Creates indexids when content is disabled.

        Args:
            ids: optional ids to add

        Returns:
            new indexids, if content disabled
        """

        return IndexIds(self, ids) if not self.config.get("content") else None

    def createindexes(self):
        """
        Creates subindexes from config.

        Returns:
            list of subindexes
        """

        if self.indexes:
            self.indexes.close()

        if "indexes" in self.config:
            indexes = {}
            for index, config in self.config["indexes"].items():
                indexes[index] = Embeddings(config, models=self.models)

            return Indexes(self, indexes)

        return None

    def createscoring(self):
        """
        Creates a scoring from config.

        Returns:
            new scoring, if enabled in config
        """

        if self.scoring:
            self.scoring.close()

        if "scoring" in self.config:
            config = self.config["scoring"]
            config = config if isinstance(config, dict) else {"method": config}

            config = self.columns(config)
            return ScoringFactory.create(config, self.models)

        return None

    def hassparse(self):
        """
        Checks is this embeddings database has an associated sparse index.

        Returns:
            True if this embeddings has an associated scoring index
        """

        return ScoringFactory.issparse(self.config.get("scoring"))

    def columns(self, config):
        """
        Adds custom text/object column information if it's provided.

        Args:
            config: input configuration

        Returns:
            config with column information added
        """

        if "columns" in self.config:
            config = config.copy()

            config["columns"] = self.config["columns"]

        return config



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from io import StringIO
from shlex import shlex

from .expression import Expression


class SQL:
    """
    Translates melowrag SQL statements into database native queries.
    """

    CLAUSES: typing.ClassVar = ["select", "from", "where", "group", "having", "order", "limit", "offset"]

    def __init__(self, database=None, tolist=False):
        """
        Creates a new SQL query parser.

        Args:
            database: database instance that provides resolver callback, if any
            tolist: outputs expression lists if True, expression text otherwise, defaults to False
        """

        self.expression = Expression(database.resolve if database else self.defaultresolve, tolist)

    def __call__(self, query):
        """
        Parses an input SQL query and normalizes column names in the query clauses. This method will also embed
        similarity search placeholders into the query.

        Args:
            query: input query

        Returns:
            {clause name: clause text}
        """

        clauses = None
        if self.issql(query):
            query = query.split(";")[0]

            tokens, positions = self.tokenize(query)

            aliases, similar = {}, []

            clauses = {
                "select": self.parse(tokens, positions, "select", alias=True, aliases=aliases),
                "where": self.parse(tokens, positions, "where", aliases=aliases, similar=similar),
                "groupby": self.parse(tokens, positions, "group", offset=2, aliases=aliases),
                "having": self.parse(tokens, positions, "having", aliases=aliases),
                "orderby": self.parse(tokens, positions, "order", offset=2, aliases=aliases),
                "limit": self.parse(tokens, positions, "limit", aliases=aliases),
                "offset": self.parse(tokens, positions, "offset", aliases=aliases),
            }

            if similar:
                clauses["similar"] = similar

        return clauses if clauses else {"similar": [[query]]}

    # pylint: disable=W0613
    def defaultresolve(self, name, alias=None):
        """
        Default resolve function. Performs no processing, only returns name.

        Args:
            name: query column name
            alias: alias name, defaults to None

        Returns:
            name
        """

        return name

    def issql(self, query):
        """
        Detects if this is a SQL query.

        Args:
            query: input query

        Returns:
            True if this is a valid SQL query, False otherwise
        """

        if isinstance(query, str):
            query = query.lower().strip(";").replace("\n", " ").replace("\t", " ").strip()

            return query.startswith("select ") and (" from melowrag " in query or query.endswith(" from melowrag"))

        return False

    def snippet(self, text):
        """
        Parses a partial SQL snippet.

        Args:
            text: SQL snippet

        Returns:
            parsed snippet
        """

        tokens, _ = self.tokenize(text)
        return self.expression(tokens)

    def tokenize(self, query):
        """
        Tokenizes SQL query into tokens.

        Args:
            query: input query

        Returns:
            (tokenized query, token positions)
        """

        tokens = shlex(StringIO(query), punctuation_chars="=!<>+-*/%|")
        tokens.wordchars += ":@#"
        tokens.commenters = ""
        tokens = list(tokens)

        positions = {}

        for x, token in enumerate(tokens):
            t = token.lower()
            if (
                t not in positions
                and t in SQL.CLAUSES
                and (t not in ["group", "order"] or (x + 1 < len(tokens) and tokens[x + 1].lower() == "by"))
            ):
                positions[t] = x

        return (tokens, positions)

    def parse(self, tokens, positions, name, offset=1, alias=False, aliases=None, similar=None):
        """
        Runs query column name to database column name mappings for clauses. This method will also
        parse SIMILAR() function calls, extract parameters for those calls and leave a placeholder
        to be filled in with similarity results.

        Args:
            tokens: query tokens
            positions: token positions - used to locate the start of sql clauses
            name: current query clause name
            offset: how many tokens are in the clause name
            alias: True if terms in the clause should be aliased (i.e. column as alias)
            aliases: dict of generated aliases, if present these tokens should NOT be resolved
            similar: list where parsed similar clauses should be stored

        Returns:
            formatted clause
        """

        clause = None
        if name in positions:
            end = [positions.get(x, len(tokens)) for x in SQL.CLAUSES[SQL.CLAUSES.index(name) + 1 :]]
            end = min(end) if end else len(tokens)

            clause = tokens[positions[name] + offset : end]

            clause = self.expression(clause, alias, aliases, similar)

        return clause


class SQLError(Exception):
    """
    Raised for errors generated by user SQL queries
    """



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
IndexIds module
"""

from ...serialization import Serializer


class IndexIds:
    """
    Stores index ids when content is disabled.
    """

    def __init__(self, embeddings, ids=None):
        """
        Creates an IndexIds instance.

        Args:
            embeddings: embeddings instance
            ids: ids to store
        """

        self.config = embeddings.config
        self.ids = ids

    def __iter__(self):
        yield from self.ids

    def __getitem__(self, index):
        return self.ids[index]

    def __setitem__(self, index, value):
        self.ids[index] = value

    def __add__(self, ids):
        return self.ids + ids

    def load(self, path):
        """
        Loads IndexIds from path.

        Args:
            path: path to load
        """

        if "ids" in self.config:
            self.ids = self.config.pop("ids")
        else:
            self.ids = Serializer.load(path)

    def save(self, path):
        """
        Saves IndexIds to path.

        Args:
            path: path to save
        """

        Serializer.save(self.ids, path)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .messagepack import MessagePack
from .pickle import Pickle


class SerializeFactory:
    """
    Methods to create data serializers.
    """

    @staticmethod
    def create(method=None, **kwargs):
        """
        Creates a new Serialize instance.

        Args:
            method: serialization method
            kwargs: additional keyword arguments to pass to serialize instance
        """

        if method == "pickle":
            return Pickle(**kwargs)

        return MessagePack(**kwargs)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from zipfile import ZIP_DEFLATED, ZipFile

from .compress import Compress


class Zip(Compress):
    """
    Zip compression
    """

    def pack(self, path, output):
        with ZipFile(output, "w", ZIP_DEFLATED) as zfile:
            for root, _, files in sorted(os.walk(path)):
                for f in files:
                    name = os.path.join(os.path.relpath(root, path), f)

                    zfile.write(os.path.join(root, f), arcname=name)

    def unpack(self, path, output):
        with ZipFile(path, "r") as zfile:
            for fullpath in zfile.namelist():
                fullpath = os.path.join(path, fullpath)
                if os.path.dirname(fullpath) and not self.validate(path, fullpath):
                    raise OSError(f"Invalid zip entry: {fullpath}")

            zfile.extractall(output)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Questions module
"""

from ..hfpipeline import HFPipeline


class Questions(HFPipeline):
    """
    Runs extractive QA for a series of questions and contexts.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        super().__init__("question-answering", path, quantize, gpu, model, **kwargs)

    def __call__(self, questions, contexts, workers=0):
        """
        Runs a extractive question-answering model against each question-context pair, finding the best answers.

        Args:
            questions: list of questions
            contexts: list of contexts to pull answers from
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            list of answers
        """

        answers = []

        for x, question in enumerate(questions):
            if question and contexts[x]:
                result = self.pipeline(question=question, context=contexts[x], num_workers=workers)

                answer, score = result["answer"], result["score"]

                if score < 0.05:
                    answer = None

                answers.append(answer)
            else:
                answers.append(None)

        return answers



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .tfidf import TFIDF


class SIF(TFIDF):
    """
    Smooth Inverse Frequency (SIF) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        self.a = self.config.get("a", 1e-3)

    def computefreq(self, tokens):
        return {token: self.wordfreq[token] for token in tokens}

    def score(self, freq, idf, length):
        if isinstance(freq, np.ndarray) and freq.shape != np.array(idf).shape:
            freq.fill(freq.sum())

        return self.a / (self.a + freq / self.tokens)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..pipelines import Tokenizer
from ..score_functions import ScoringFactory


class Topics:
    """
    Topic modeling using community detection.
    """

    def __init__(self, config):
        """
        Creates a new Topics instance.

        Args:
            config: topic configuration
        """

        self.config = config if config else {}
        self.tokenizer = Tokenizer(stopwords=True)

        self.stopwords = set()
        if "stopwords" in self.config:
            self.stopwords.update(self.config["stopwords"])

    def __call__(self, graph):
        """
        Runs topic modeling for input graph.

        Args:
            graph: Graph instance

        Returns:
            dictionary of {topic name: [ids]}
        """

        communities = graph.communities(self.config)

        communities = sorted(communities, key=len, reverse=True)

        centrality = graph.centrality()

        topics = [self.score(graph, x, community, centrality) for x, community in enumerate(communities)]

        return self.merge(topics)

    def score(self, graph, index, community, centrality):
        """
        Scores a community of nodes and generates the topn terms in the community.

        Args:
            graph: Graph instance
            index: community index
            community: community of nodes
            centrality: node centrality scores

        Returns:
            (topn topic terms, topic ids sorted by score descending)
        """

        scoring = ScoringFactory.create({"method": self.config.get("labels", "bm25"), "terms": True})
        scoring.index((node, self.tokenize(graph, node), None) for node in community)

        if scoring.idf:
            idf = sorted(scoring.idf, key=scoring.idf.get)

            topn = self.config.get("terms", 4)

            terms = self.topn(idf, topn)

            community = [uid for uid, _ in scoring.search(terms, len(community))]
        else:
            terms = ["topic", str(index)]

            community = sorted(community, key=lambda x: centrality[x], reverse=True)

        return (terms, community)

    def tokenize(self, graph, node):
        """
        Tokenizes node text.

        Args:
            graph: Graph instance
            node: node id

        Returns:
            list of node tokens
        """

        text = graph.attribute(node, "text")
        return self.tokenizer(text) if text else []

    def topn(self, terms, n):
        """
        Gets topn terms.

        Args:
            terms: list of terms
            n: topn

        Returns:
            topn terms
        """

        topn = []

        for term in terms:
            if self.tokenizer(term) and term not in self.stopwords:
                topn.append(term)

            if len(topn) == n:
                break

        return topn

    def merge(self, topics):
        """
        Merges duplicate topics

        Args:
            topics: list of (topn terms, topic ids)

        Returns:
            dictionary of {topic name:[ids]}
        """

        merge, termslist = {}, {}

        for terms, uids in topics:
            key = frozenset(terms)

            if key not in merge:
                merge[key], termslist[key] = [], terms

            merge[key].extend(uids)

        results = {}
        for k, v in sorted(merge.items(), key=lambda x: len(x[1]), reverse=True):
            results["_".join(termslist[k])] = v

        return results



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utilities import Resolver
from .sbert import SparseSTVectors


class SparseVectorsFactory:
    """
    Methods to create sparse vector models.
    """

    @staticmethod
    def create(config, models=None):
        """
        Create a Vectors model instance.

        Args:
            config: vector configuration
            models: models cache

        Returns:
            Vectors
        """

        method = config.get("method", "sentence-transformers")

        if method == "sentence-transformers":
            return SparseSTVectors(config, None, models) if config and config.get("path") else None

        return SparseVectorsFactory.resolve(method, config, models) if method else None

    @staticmethod
    def resolve(backend, config, models):
        """
        Attempt to resolve a custom backend.

        Args:
            backend: backend class
            config: vector configuration
            models: models cache

        Returns:
            Vectors
        """

        try:
            return Resolver()(backend)(config, None, models)
        except Exception as e:
            raise ImportError(f"Unable to resolve sparse vectors backend: '{backend}'") from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from .base import Task


class UrlTask(Task):
    """
    Task that processes urls
    """

    PREFIX = r"\w+:\/\/"

    def accept(self, element):
        return super().accept(element) and re.match(UrlTask.PREFIX, element.lower())



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base_compression import Archive


class ArchiveFactory:
    """
    Methods to create Archive instances.
    """

    @staticmethod
    def create(directory=None):
        """
        Create a new Archive instance.

        Args:
            directory: optional default working directory, otherwise uses a temporary directory

        Returns:
            Archive
        """

        return Archive(directory)

    def __repr__(self):
        """
        Returns a detailed string representation of the ArchiveFactory class.
        """
        return f"{self.__class__.__name__}(factory for Archive instances)"

    def __str__(self):
        """
        Returns a user-friendly string representation of the ArchiveFactory class.
        """
        return "ArchiveFactory: use .create(directory) to get an Archive instance"



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .rdbms import RDBMS


class Embedded(RDBMS):
    """
    Base class for embedded relational databases. An embedded relational database stores all content in a local file.
    """

    def __init__(self, config):
        """
        Creates a new Database.

        Args:
            config: database configuration parameters
        """

        super().__init__(config)

        self.path = None

    def load(self, path):
        super().load(path)

        self.path = path

    def save(self, path):
        if not self.path:
            self.connection.commit()

            connection = self.copy(path)

            self.connection.close()

            self.session(connection=connection)
            self.path = path

        elif self.path == path:
            self.connection.commit()

        else:
            self.copy(path).close()

    def jsonprefix(self):
        return "json_extract(data"

    def jsoncolumn(self, name):
        return f"json_extract(data, '$.{name}')"

    def copy(self, path):
        """
        Copies the current database into path.

        Args:
            path: path to write database

        Returns:
            new connection with data copied over
        """

        raise NotImplementedError



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np

try:
    from scipy.sparse import vstack
    from sklearn.cluster import MiniBatchKMeans
    from sklearn.metrics import pairwise_distances_argmin_min
    from sklearn.utils.extmath import safe_sparse_dot

    IVFSPARSE = True
except ImportError:
    IVFSPARSE = False

from ...serialization import SerializeFactory
from ...utilities import SparseArray
from ..base import VectoreIndex


class IVFSparse(VectoreIndex):
    """
    Inverted file (IVF) index with flat vector file storage and sparse array support.

    IVFSparse builds an IVF index and enables approximate nearest neighbor (VectoreIndex) search.

    This index is modeled after Faiss and supports many of the same parameters.

    See this link for more: https://github.com/facebookresearch/faiss/wiki/Faster-search
    """

    def __init__(self, config):
        super().__init__(config)

        if not IVFSPARSE:
            raise ImportError('IVFSparse is not available - install "ann" extra to enable')

        self.centroids = None

        self.ids = None

        self.blocks = None

        self.deletes = None

    def index(self, embeddings):
        train, sample = embeddings, self.setting("sample")
        if sample:
            rng = np.random.default_rng(0)
            indices = sorted(
                rng.choice(
                    train.shape[0],
                    int(sample * train.shape[0]),
                    replace=False,
                    shuffle=False,
                )
            )
            train = train[indices]

        clusters = self.nlist(embeddings.shape[0], train.shape[0])

        if clusters > 1:
            kmeans = MiniBatchKMeans(n_clusters=clusters, random_state=0, n_init="auto").fit(train)

            indices, _ = pairwise_distances_argmin_min(kmeans.cluster_centers_, train, metric="cosine")
            self.centroids = embeddings[np.unique(indices)]

        ids = self.aggregate(embeddings)

        self.ids = dict(sorted(ids.items(), key=lambda x: x[0]))

        self.blocks = {k: embeddings[v] for k, v in self.ids.items()}

        self.deletes = []

        self.config["offset"] = embeddings.shape[0]
        self.metadata({"clusters": clusters})

    def append(self, embeddings):
        offset = self.size()

        for cluster, ids in self.aggregate(embeddings).items():
            self.ids[cluster].extend([x + offset for x in ids])

            self.blocks[cluster] = vstack([self.blocks[cluster], embeddings[ids]])

        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        self.deletes.extend(ids)

    def search(self, queries, limit):
        results = []
        for query in queries:
            if self.centroids is not None:
                indices, _ = self.topn(query, self.centroids, self.nprobe())

                ids = np.concatenate([self.ids[x] for x in indices if x in self.ids])

                data = vstack([self.blocks[x] for x in indices if x in self.blocks])
            else:
                ids, data = np.array(self.ids[0]), self.blocks[0]

            deletes = np.argwhere(np.isin(ids, self.deletes)).ravel()

            indices, scores = self.topn(query, data, limit, deletes)

            results.append(list(zip(ids[indices].tolist(), scores.tolist(), strict=False)))

        return results

    def count(self):
        return self.size() - len(self.deletes)

    def load(self, path):
        serializer = SerializeFactory.create("msgpack", streaming=True, read_size=1)

        with open(path, "rb") as f:
            unpacker = serializer.loadstream(f)
            header = next(unpacker)

            self.centroids = SparseArray().load(f) if header["centroids"] else None

            self.ids = dict(next(unpacker))

            self.blocks = {}
            for key in self.ids:
                self.blocks[key] = SparseArray().load(f)

            self.deletes = next(unpacker)

    def save(self, path):
        serializer = SerializeFactory.create("msgpack")

        with open(path, "wb") as f:
            serializer.savestream(
                {
                    "centroids": self.centroids is not None,
                    "count": self.count(),
                    "blocks": len(self.blocks),
                },
                f,
            )

            if self.centroids is not None:
                SparseArray().save(f, self.centroids)

            serializer.savestream(list(self.ids.items()), f)

            for block in self.blocks.values():
                SparseArray().save(f, block)

            serializer.savestream(self.deletes, f)

    def aggregate(self, data):
        """
        Aggregates input data array into clusters. This method sorts each data element into the
        cluster with the highest cosine similarity centroid.

        Args:
            data: input data

        Returns:
            {cluster, ids}
        """

        if self.centroids is None:
            return {0: list(range(data.shape[0]))}

        indices, _ = pairwise_distances_argmin_min(data, self.centroids, metric="cosine")

        ids = {}
        for x, cluster in enumerate(indices.tolist()):
            if cluster not in ids:
                ids[cluster] = []

            ids[cluster].append(x)

        return ids

    def topn(self, query, data, limit, deletes=None):
        """
        Gets the top n most similar data elements for query.

        Args:
            query: input query array
            data: data array
            limit: top n
            deletes: optional list of deletes to filter from results

        Returns:
            list of matching (indices, scores)
        """

        scores = safe_sparse_dot(query, data.T, dense_output=True)

        if deletes is not None:
            scores[:, deletes] = 0

        indices = np.argpartition(-scores, limit if limit < scores.shape[0] else scores.shape[0] - 1)[:, :limit]
        scores = np.clip(np.take_along_axis(scores, indices, axis=1), 0.0, 1.0)

        return indices[0], scores[0]

    def nlist(self, count, train):
        """
        Calculates the number of clusters for this IVFSparse index. Note that the final number of clusters
        could be lower as duplicate cluster centroids are filtered out.

        Args:
            count: initial dataset size
            train: number of rows used to train

        Returns:
            number of clusters
        """

        default = 1 if count <= 5000 else self.cells(train)

        return self.setting("nlist", default)

    def nprobe(self):
        """
        Gets or derives the nprobe search parameter.

        Returns:
            nprobe setting
        """

        size = self.size()

        default = 6 if size <= 5000 else self.cells(size) // 64
        return self.setting("nprobe", default)

    def cells(self, count):
        """
        Calculates the number of IVF cells for an IVFSparse index.

        Args:
            count: number of rows

        Returns:
            number of IVF cells
        """

        return max(min(round(4 * math.sqrt(count)), int(count / 39)), 1)

    def size(self):
        """
        Gets the total size of this index including deletes.

        Returns:
            size
        """

        return sum(len(x) for x in self.ids.values())



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Database imports
"""

from .base import Database
from .client import Client
from .db_encoder import Encoder, EncoderFactory, ImageEncoder, SerializeEncoder
from .db_schema import Base, Batch, Document, Object, Score, Section, SectionBase, Statement, idcolumn
from .db_sql import SQL, Aggregate, Expression, SQLError, Token
from .duckdb import DuckDB
from .embedded import Embedded
from .factory import DatabaseFactory
from .rdbms import RDBMS
from .sqlite import SQLite

__all__ = (
    "RDBMS",
    "SQL",
    "Aggregate",
    "Base",
    "Batch",
    "Client",
    "Database",
    "DatabaseFactory",
    "Document",
    "DuckDB",
    "Embedded",
    "Encoder",
    "EncoderFactory",
    "Expression",
    "ImageEncoder",
    "Object",
    "SQLError",
    "SQLite",
    "Score",
    "Section",
    "SectionBase",
    "SerializeEncoder",
    "Statement",
    "Token",
    "idcolumn",
)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Resolver:
    """
    Resolves a Python class path
    """

    def __call__(self, path):
        """
        Class instance to resolve.

        Args:
            path: path to class

        Returns:
            class instance
        """

        parts = path.split(".")

        module = ".".join(parts[:-1])
        m = __import__(module)
        for comp in parts[1:]:
            m = getattr(m, comp)

        return m



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Entity module
"""

try:
    from gliner import GLiNER

    GLINER = True
except ImportError:
    GLINER = False

from huggingface_hub.errors import HFValidationError
from transformers.utils import cached_file

from ...modeling import Models
from ..hfpipeline import HFPipeline


class Entity(HFPipeline):
    """
    Applies a token classifier to text and extracts entity/label combinations.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        self.gliner = self.isgliner(path)
        if self.gliner:
            if not GLINER:
                raise ImportError('GLiNER is not available - install "pipeline" extra to enable')

            self.pipeline = GLiNER.from_pretrained(path)
            self.pipeline = self.pipeline.to(Models.device(Models.deviceid(gpu)))
        else:
            super().__init__("token-classification", path, quantize, gpu, model, **kwargs)

    def __call__(self, text, labels=None, aggregate="simple", flatten=None, join=False, workers=0):
        """
        Applies a token classifier to text and extracts entity/label combinations.

        Args:
            text: text|list
            labels: list of entity type labels to accept, defaults to None which accepts all
            aggregate: method to combine multi token entities - options are "simple" (default), "first", "average" or "max"
            flatten: flatten output to a list of labels if present. Accepts a boolean or float value to only keep scores greater than that number.
            join: joins flattened output into a string if True, ignored if flatten not set
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            list of (entity, entity type, score) or list of entities depending on flatten parameter
        """

        results = self.execute(text, labels, aggregate, workers)

        if isinstance(text, str):
            results = [results]

        threshold = 0.0 if isinstance(flatten, bool) else flatten

        outputs = []
        for result in results:
            if flatten:
                output = [
                    r["word"] for r in result if self.accept(r["entity_group"], labels) and r["score"] >= threshold
                ]
                outputs.append(" ".join(output) if join else output)
            else:
                outputs.append(
                    [
                        (r["word"], r["entity_group"], float(r["score"]))
                        for r in result
                        if self.accept(r["entity_group"], labels)
                    ]
                )

        return outputs[0] if isinstance(text, str) else outputs

    def isgliner(self, path):
        """
        Tests if path is a GLiNER model.

        Args:
            path: model path

        Returns:
            True if this is a GLiNER model, False otherwise
        """

        try:
            return cached_file(path_or_repo_id=path, filename="gliner_config.json") is not None

        except (HFValidationError, OSError):
            pass

        return False

    def execute(self, text, labels, aggregate, workers):
        """
        Runs the entity extraction pipeline.

        Args:
            text: text|list
            labels: list of entity type labels to accept, defaults to None which accepts all
            aggregate: method to combine multi token entities - options are "simple" (default), "first", "average" or "max"
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            list of entities and labels
        """

        if self.gliner:
            results = self.pipeline.batch_predict_entities(
                text if isinstance(text, list) else [text], labels if labels else ["person", "organization", "location"]
            )

            entities = []
            for result in results:
                entities.append([{"word": x["text"], "entity_group": x["label"], "score": x["score"]} for x in result])

            return entities if isinstance(text, list) else entities[0]

        return self.pipeline(text, aggregation_strategy=aggregate, num_workers=workers)

    def accept(self, etype, labels):
        """
        Determines if entity type is in valid entity type.

        Args:
            etype: entity type
            labels: list of entities to accept

        Returns:
            if etype is accepted
        """

        return not labels or etype in labels



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import tempfile
import uuid

import numpy as np

from ..pipelines import Tokenizer
from .recovery import Recovery


class Vectors:
    """
    Base class for vector models. Vector models transform input content into numeric vectors.
    """

    def __init__(self, config, scoring, models):
        """
        Creates a new vectors instance.

        Args:
            config: vector configuration
            scoring: optional scoring instance for term weighting
            models: models cache
        """

        self.config = config
        self.scoring = scoring
        self.models = models

        if config:
            self.initialized = "dimensions" in config

            self.tokenize = config.get("tokenize")

            self.model = self.load(config.get("path"))

            self.encodebatch = config.get("encodebatch", 32)

            self.instructions = config.get("instructions")

            self.dimensionality = config.get("dimensionality")

            quantize = config.get("quantize")
            self.qbits = (
                max(min(quantize, 8), 1) if isinstance(quantize, int) and not isinstance(quantize, bool) else None
            )

    def loadmodel(self, path):
        """
        Loads vector model at path.

        Args:
            path: path to vector model

        Returns:
            vector model
        """

        raise NotImplementedError

    def encode(self, data, category=None):
        """
        Encodes a batch of data using vector model.

        Args:
            data: batch of data
            category: optional category for instruction-based embeddings

        Return:
            transformed data
        """

        raise NotImplementedError

    def load(self, path):
        """
        Loads a model using the current configuration. This method will return previously cached models
        if available.

        Returns:
            model
        """

        if self.models and path in self.models:
            return self.models[path]

        model = self.loadmodel(path)

        if self.models is not None and path:
            self.models[path] = model

        return model

    def index(self, documents, batchsize=500, checkpoint=None):
        """
        Converts a list of documents to a temporary file with embeddings arrays. Returns a tuple of document ids,
        number of dimensions and temporary file with embeddings.

        Args:
            documents: list of (id, data, tags)
            batchsize: index batch size
            checkpoint: optional checkpoint directory, enables indexing restart

        Returns:
            (ids, dimensions, batches, stream)
        """

        ids, dimensions, batches, stream = [], None, 0, None

        vectorsid = self.vectorsid() if checkpoint else None
        recovery = Recovery(checkpoint, vectorsid, self.loadembeddings) if checkpoint else None

        with self.spool(checkpoint, vectorsid) as output:
            stream = output.name
            batch = []
            for document in documents:
                batch.append(document)

                if len(batch) == batchsize:
                    uids, dimensions = self.batch(batch, output, recovery)
                    ids.extend(uids)
                    batches += 1

                    batch = []

            if batch:
                uids, dimensions = self.batch(batch, output, recovery)
                ids.extend(uids)
                batches += 1

        return (ids, dimensions, batches, stream)

    def vectors(self, documents, batchsize=500, checkpoint=None, buffer=None, dtype=None):
        """
        Bulk encodes documents into vectors using index(). Return the data as a mmap-ed array.

        Args:
            documents: list of (id, data, tags)
            batchsize: index batch size
            checkpoint: optional checkpoint directory, enables indexing restart
            buffer: file path used for memmap buffer
            dtype: dtype for buffer

        Returns:
            (ids, dimensions, embeddings)
        """

        ids, dimensions, batches, stream = self.index(documents, batchsize, checkpoint)

        embeddings = None
        if ids:
            embeddings = np.memmap(buffer, dtype=dtype, shape=(len(ids), dimensions), mode="w+")
            with open(stream, "rb") as queue:
                x = 0
                for _ in range(batches):
                    batch = self.loadembeddings(queue)
                    embeddings[x : x + batch.shape[0]] = batch
                    x += batch.shape[0]

        if not checkpoint:
            os.remove(stream)

        return (ids, dimensions, embeddings)

    def close(self):
        """
        Closes this vectors instance.
        """

        self.model = None

    def transform(self, document):
        """
        Transforms document into an embeddings vector.

        Args:
            document: (id, data, tags)

        Returns:
            embeddings vector
        """

        return self.batchtransform([document])[0]

    def batchtransform(self, documents, category=None):
        """
        Transforms batch of documents into embeddings vectors.

        Args:
            documents: list of documents used to build embeddings
            category: category for instruction-based embeddings

        Returns:
            embeddings vectors
        """

        documents = [self.prepare(data, category) for _, data, _ in documents]

        if documents and isinstance(documents[0], np.ndarray):
            return np.array(documents, dtype=np.float32)

        return self.vectorize(documents, category)

    def dot(self, queries, data):
        """
        Calculates the dot product similarity between queries and documents. This method
        assumes each of the inputs are normalized.

        Args:
            queries: queries
            data: search data

        Returns:
            dot product scores
        """

        return np.dot(queries, data.T).tolist()

    def vectorsid(self):
        """
        Generates vectors uid for this vectors instance.

        Returns:
            vectors uid
        """

        select = ["path", "method", "tokenizer", "maxlength", "tokenize", "instructions", "dimensionality", "quantize"]
        config = {k: v for k, v in self.config.items() if k in select}
        config.update(self.config.get("vectors", {}))

        return str(uuid.uuid5(uuid.NAMESPACE_DNS, json.dumps(config, sort_keys=True)))

    def spool(self, checkpoint, vectorsid):
        """
        Opens a spool file for queuing generated vectors.

        Args:
            checkpoint: optional checkpoint directory, enables indexing restart
            vectorsid: vectors uid for current configuration

        Returns:
            vectors spool file
        """

        if checkpoint:
            os.makedirs(checkpoint, exist_ok=True)
            return open(f"{checkpoint}/{vectorsid}", "wb")

        return tempfile.NamedTemporaryFile(mode="wb", suffix=".npy", delete=False)

    def batch(self, documents, output, recovery):
        """
        Builds a batch of embeddings.

        Args:
            documents: list of documents used to build embeddings
            output: output temp file to store embeddings
            recovery: optional recovery instance

        Returns:
            (ids, dimensions) list of ids and number of dimensions in embeddings
        """

        ids = [uid for uid, _, _ in documents]
        documents = [self.prepare(data, "data") for _, data, _ in documents]
        dimensions = None

        embeddings = recovery() if recovery else None
        embeddings = self.vectorize(documents, "data") if embeddings is None else embeddings
        if embeddings is not None:
            dimensions = embeddings.shape[1]
            self.saveembeddings(output, embeddings)

        return (ids, dimensions)

    def prepare(self, data, category=None):
        """
        Prepares input data for vector model.

        Args:
            data: input data
            category: category for instruction-based embeddings

        Returns:
            data formatted for vector model
        """

        data = self.tokens(data)

        category = category if category else "query"

        if self.instructions and category in self.instructions and isinstance(data, str):
            data = f"{self.instructions[category]}{data}"

        return data

    def tokens(self, data):
        """
        Prepare data as tokens model can accept.

        Args:
            data: input data

        Returns:
            tokens formatted for model
        """

        if self.tokenize and isinstance(data, str):
            data = Tokenizer.tokenize(data)

        if isinstance(data, list):
            data = " ".join(data)

        return data

    def vectorize(self, data, category=None):
        """
        Runs data vectorization, which consists of the following steps.

          1. Encode data into vectors using underlying model
          2. Truncate vectors, if necessary
          3. Normalize vectors
          4. Quantize vectors, if necessary

        Args:
            data: input data
            category: category for instruction-based embeddings

        Returns:
            embeddings vectors
        """

        category = category if category else "query"

        embeddings = self.encode(data, category)

        if embeddings is not None:
            if self.dimensionality and self.dimensionality < embeddings.shape[1]:
                embeddings = self.truncate(embeddings)

            embeddings = self.normalize(embeddings)

            if self.qbits:
                embeddings = self.quantize(embeddings)

        return embeddings

    def loadembeddings(self, f):
        """
        Loads embeddings from file.

        Args:
            f: file to load from

        Returns:
            embeddings
        """

        return np.load(f, allow_pickle=False)

    def saveembeddings(self, f, embeddings):
        """
        Saves embeddings to output.

        Args:
            f: output file
            embeddings: embeddings to save
        """

        np.save(f, embeddings, allow_pickle=False)

    def truncate(self, embeddings):
        """
        Truncates embeddings to the configured dimensionality.

        This is only useful for models trained to store more important information in
        earlier dimensions such as Matryoshka Representation Learning (MRL).

        Args:
            embeddings: input embeddings

        Returns:
            truncated embeddings
        """

        return embeddings[:, : self.dimensionality]

    def normalize(self, embeddings):
        """
        Normalizes embeddings using L2 normalization. Operation applied directly on array.

        Args:
            embeddings: input embeddings

        Returns:
            embeddings
        """

        if len(embeddings.shape) > 1:
            embeddings /= np.linalg.norm(embeddings, axis=1)[:, np.newaxis]
        else:
            embeddings /= np.linalg.norm(embeddings)

        return embeddings

    def quantize(self, embeddings):
        """
        Quantizes embeddings using scalar quantization.

        Args:
            embeddings: input embeddings

        Returns:
            quantized embeddings
        """

        factor = 2 ** (self.qbits - 1)

        scalars = embeddings * factor
        scalars = scalars.clip(-factor, factor - 1) + factor
        scalars = scalars.astype(np.uint8)

        bits = np.unpackbits(scalars.reshape(-1, 1), axis=1)

        bits = bits[:, -self.qbits :]

        return np.packbits(bits.reshape(embeddings.shape[0], embeddings.shape[1] * self.qbits), axis=1)

    def __repr__(self):
        """
        Returns a detailed string representation of the Vectors instance.

        Returns:
            Detailed string representation
        """
        class_name = self.__class__.__name__
        config_info = {}
        if hasattr(self, "config") and self.config:
            config_info.update(
                {
                    "path": self.config.get("path"),
                    "method": self.config.get("method"),
                    "tokenize": getattr(self, "tokenize", None),
                    "encodebatch": getattr(self, "encodebatch", None),
                    "dimensionality": getattr(self, "dimensionality", None),
                    "quantize": getattr(self, "qbits", None),
                    "instructions": bool(getattr(self, "instructions", None)),
                }
            )
        config_info = {k: v for k, v in config_info.items() if v is not None}
        status = "initialized" if getattr(self, "initialized", False) else "uninitialized"
        return f"{class_name}({config_info}, status={status})"

    def __str__(self):
        """
        Returns a user-friendly string representation of the Vectors instance.

        Returns:
            User-friendly string representation
        """
        class_name = self.__class__.__name__
        details = []
        if hasattr(self, "config") and self.config:
            if self.config.get("path"):
                details.append(f"model_path='{self.config['path']}'")
            if self.config.get("method"):
                details.append(f"method='{self.config['method']}'")
        if hasattr(self, "dimensionality") and self.dimensionality:
            details.append(f"dimensions={self.dimensionality}")
        if hasattr(self, "encodebatch") and self.encodebatch:
            details.append(f"batch_size={self.encodebatch}")
        if hasattr(self, "tokenize") and self.tokenize:
            details.append("tokenized")
        if hasattr(self, "qbits") and self.qbits:
            details.append(f"quantized({self.qbits}bits)")
        if hasattr(self, "instructions") and self.instructions:
            details.append("instruction-based")
        status = "✓" if getattr(self, "initialized", False) else "✗"
        details_str = ", ".join(details) if details else "default"
        return f"{class_name}({details_str}) [{status}]"
